import pdb
import torch
import torch.nn as nn
from torch.cuda.amp import autocast
from torch.utils.data import DataLoader
from pytrial.utils.trainer import Trainer
from typing import List, Dict, Tuple, Iterable, Type, Union, Callable, Optional

class SeqSimTrainer(Trainer):
    '''
    Subclass the original trainer and provide specific evaluation functions and test_dataloader for sequence simulation.
    '''
    pass

class SeqSimEvaTrainer(Trainer):
    '''
    Subclass the original trainer and provide specific evaluation functions and test_dataloader for sequence simulation.
    '''
    def prepare_input(self, data):
        '''
        Prepare inputs for sequence simulation models.

        Parameters
        ----------
        data: dict[list]
            A single patient records.

        idx: int
            The patient index.

        vdx: int
            The target visit index.

        '''
        return self.model._prepare_input(data)
    
    def train_one_iteration(self, 
        max_grad_norm=None,
        warmup_steps=None,
        use_amp=None, 
        scaler=None,
        train_loss_dict=None):
        
        skip_scheduler = False
        num_train_objectives = len(self.train_dataloader)
        for train_idx in range(num_train_objectives):
            data_iterator = self.data_iterators[train_idx]
            loss_model = self.loss_models[train_idx]
            loss_model.zero_grad()
            loss_model.train()
            optimizer = self.optimizers[train_idx]
            scheduler = self.schedulers[train_idx]
            try:
                data = next(data_iterator)
            except StopIteration:
                data_iterator = iter(self.train_dataloader[train_idx])
                self.data_iterators[train_idx] = data_iterator
                data = next(data_iterator)
                
            
            inputs = self.prepare_input(data)
            if use_amp:
                with autocast():
                    loss_return = loss_model(inputs)
                    loss = loss_return['loss_value']
            else:
                loss_return = loss_model(inputs)
                loss = loss_return['loss_value']

            # update for a single sample's all visits
            if use_amp:
                scale_before_step = scaler.get_scale()
                scaler.scale(loss).backward()
                scaler.unscale_(optimizer)
                torch.nn.utils.clip_grad_norm_(loss_model.parameters(), max_grad_norm)
                scaler.step(optimizer)
                scaler.update()
                skip_scheduler = scaler.get_scale() != scale_before_step
            else:
                loss.backward()
                torch.nn.utils.clip_grad_norm_(loss_model.parameters(), max_grad_norm)
                optimizer.step()

            train_loss_dict[train_idx].append(loss.item())
        
        if use_amp:
            skip_scheduler = scaler.get_scale() != scale_before_step

        if not skip_scheduler and warmup_steps > 0:
            scheduler.step()

    def _build_optimizer(self,
        loss_models,
        warmup_steps,
        num_train_steps,
        optimizer_param,
        weight_decay,
        scheduler=None,
        ):
        optimizer_class = torch.optim.AdamW
        optimizers = []
        schedulers = []
        for loss_model in loss_models:
            param_optimizer = list(loss_model.named_parameters())
            no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
            optimizer_grouped_parameters = [
                {'params': [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)], 'weight_decay': weight_decay},
                {'params': [p for n, p in param_optimizer if any(nd in n for nd in no_decay)], 'weight_decay': 0.0}
            ]
            optimizer = optimizer_class(optimizer_grouped_parameters, **optimizer_param)
            if warmup_steps > 0:
                scheduler_obj = self._build_scheduler(optimizer, scheduler, warmup_steps=warmup_steps, t_total=num_train_steps)
            else:
                scheduler_obj = None
            optimizers.append(optimizer)
            schedulers.append(scheduler_obj)
        return optimizers, schedulers
    
class SeqSimSyntegTrainer(Trainer):
    '''
    Subclass the original trainer and provide specific evaluation functions and test_dataloader for sequence simulation.
    '''
    def __init__(self, model: nn.Module, train_objectives: List[Tuple[DataLoader, nn.Module]], condition=False, test_data=None, test_metric=None, less_is_better=False, load_best_at_end=True, n_gpus=1, output_dir='./checkpoints/', **kwargs):
        super().__init__(model, train_objectives, test_data, test_metric, less_is_better, load_best_at_end, n_gpus, output_dir, **kwargs)
        self.condition = condition
    
    def prepare_input(self, data):
        '''
        Prepare inputs for sequence simulation models.

        Parameters
        ----------
        data: dict[list]
            A single patient records.

        idx: int
            The patient index.

        vdx: int
            The target visit index.

        '''
        return self.model._prepare_input(data)
    
    def add_condition(self, data):
        return self.model._add_condition(data)
    
    def train_one_iteration(self, 
        max_grad_norm=None,
        warmup_steps=None,
        use_amp=None, 
        scaler=None,
        train_loss_dict=None):
        
        skip_scheduler = False
        num_train_objectives = len(self.train_dataloader)
        for train_idx in range(num_train_objectives):
            data_iterator = self.data_iterators[train_idx]
            loss_model = self.loss_models[train_idx]
            loss_model.zero_grad()
            loss_model.train()
            optimizer = self.optimizers[train_idx]
            scheduler = self.schedulers[train_idx]
            try:
                data = next(data_iterator)
            except StopIteration:
                data_iterator = iter(self.train_dataloader[train_idx])
                self.data_iterators[train_idx] = data_iterator
                data = next(data_iterator)
                
            
            inputs = self.prepare_input(data)
            if self.condition:
                inputs = self.add_condition(inputs)
            if use_amp:
                with autocast():
                    loss_return = loss_model(inputs)
                    loss = loss_return['loss_value']
            else:
                loss_return = loss_model(inputs)
                loss = loss_return['loss_value']

            # update for a single sample's all visits
            if use_amp:
                scale_before_step = scaler.get_scale()
                scaler.scale(loss).backward()
                scaler.unscale_(optimizer)
                torch.nn.utils.clip_grad_norm_(loss_model.parameters(), max_grad_norm)
                scaler.step(optimizer)
                scaler.update()
                skip_scheduler = scaler.get_scale() != scale_before_step
            else:
                loss.backward()
                torch.nn.utils.clip_grad_norm_(loss_model.parameters(), max_grad_norm)
                optimizer.step()

            train_loss_dict[train_idx].append(loss.item())
        
        if use_amp:
            skip_scheduler = scaler.get_scale() != scale_before_step

        if not skip_scheduler and warmup_steps > 0:
            scheduler.step()

    def _build_optimizer(self,
        loss_models,
        warmup_steps,
        num_train_steps,
        optimizer_param,
        weight_decay,
        scheduler=None,
        ):
        optimizer_class = torch.optim.AdamW
        optimizers = []
        schedulers = []
        for loss_model in loss_models:
            param_optimizer = list(loss_model.named_parameters())
            no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
            optimizer_grouped_parameters = [
                {'params': [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)], 'weight_decay': weight_decay},
                {'params': [p for n, p in param_optimizer if any(nd in n for nd in no_decay)], 'weight_decay': 0.0}
            ]
            optimizer = optimizer_class(optimizer_grouped_parameters, **optimizer_param)
            if warmup_steps > 0:
                scheduler_obj = self._build_scheduler(optimizer, scheduler, warmup_steps=warmup_steps, t_total=num_train_steps)
            else:
                scheduler_obj = None
            optimizers.append(optimizer)
            schedulers.append(scheduler_obj)
        return optimizers, schedulers

class SeqSimGANTrainer(Trainer):
    '''
    Subclass the original trainer and provide specific evaluation functions and test_dataloader for sequence simulation.
    '''
    def prepare_input(self, data, idx, vdx):
        '''
        Prepare inputs for sequence simulation models.

        Parameters
        ----------
        data: dict[list]
            A single patient records.

        idx: int
            The patient index.

        vdx: int
            The target visit index.

        '''
        return self.model._prepare_input(data, idx, vdx)
    
    def train_one_iteration(self, 
        max_grad_norm=None,
        warmup_steps=None,
        use_amp=None, 
        scaler=None,
        train_loss_dict=None):
        '''
        Default training one iteration steps, can be subclass can reimplemented.
        '''
        skip_scheduler = False
        num_train_objectives = len(self.train_dataloader)
        for train_idx in range(num_train_objectives):
            data_iterator = self.data_iterators[train_idx]
            loss_model = self.loss_models[train_idx]
            loss_model.zero_grad()
            loss_model.train()
            optimizer = self.optimizers[train_idx]
            scheduler = self.schedulers[train_idx]
            try:
                data = next(data_iterator)
            except StopIteration:
                data_iterator = iter(self.train_dataloader[train_idx])
                self.data_iterators[train_idx] = data_iterator
                data = next(data_iterator)

            # go through each single patient take the next visit as the target
            for idx, _ in enumerate(data['x']):
                num_visit = self.model._get_num_visit(data, idx)

                if num_visit < 2: # ignore patient with less than two visits
                    continue

                # go through each visit
                loss = 0 # for each sample
                for vdx in range(1, num_visit):
                    inputs = self.prepare_input(data, idx, vdx)
                    if use_amp:
                        with autocast():
                            loss_return = loss_model(inputs)
                            loss_value = loss_return['loss_value']
                    else:
                        loss_return = loss_model(inputs)
                        loss_value = loss_return['loss_value']
                    loss += loss_value

                # update for a single sample's all visits
                if use_amp:
                    scale_before_step = scaler.get_scale()
                    scaler.scale(loss).backward()
                    scaler.unscale_(optimizer)
                    torch.nn.utils.clip_grad_norm_(loss_model.parameters(), max_grad_norm)
                    scaler.step(optimizer)
                    scaler.update()
                    skip_scheduler = scaler.get_scale() != scale_before_step
                else:
                    loss.backward()
                    torch.nn.utils.clip_grad_norm_(loss_model.parameters(), max_grad_norm)
                    optimizer.step()

            train_loss_dict[train_idx].append(loss.item())
        
        if use_amp:
            skip_scheduler = scaler.get_scale() != scale_before_step

        if not skip_scheduler and warmup_steps > 0:
            scheduler.step()

    def _build_optimizer(self,
        loss_models,
        warmup_steps,
        num_train_steps,
        optimizer_param,
        weight_decay,
        scheduler=None,
        ):
        optimizer_class = torch.optim.AdamW
        optimizers = []
        schedulers = []
        for loss_model in loss_models:
            param_optimizer = list(loss_model.named_parameters())
            no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']

            if loss_model.name == 'discriminator_loss':
                # print([n for n, p in param_optimizer if (not any(nd in n for nd in no_decay)) and ('dis_module' in n)])
                optimizer_grouped_parameters = [
                    {'params': [p for n, p in param_optimizer if (not any(nd in n for nd in no_decay)) and ('dis_module' in n)], 'weight_decay': weight_decay},
                    {'params': [p for n, p in param_optimizer if any(nd in n for nd in no_decay)], 'weight_decay': 0.0}
                ]
            
            else:
                # build optimizer for generator
                optimizer_grouped_parameters = [
                    {'params': [p for n, p in param_optimizer if (not any(nd in n for nd in no_decay)) and ('dis_module' not in n)], 'weight_decay': weight_decay},
                    {'params': [p for n, p in param_optimizer if any(nd in n for nd in no_decay)], 'weight_decay': 0.0}
                ]
                # print([n for n, p in param_optimizer if (not any(nd in n for nd in no_decay)) and ('dis_module' not in n)])

            optimizer = optimizer_class(optimizer_grouped_parameters, **optimizer_param)
            if warmup_steps > 0:
                scheduler_obj = self._build_scheduler(optimizer, scheduler, warmup_steps=warmup_steps, t_total=num_train_steps)
            else:
                scheduler_obj = None
            optimizers.append(optimizer)
            schedulers.append(scheduler_obj)
        
        return optimizers, schedulers


import pdb
import os
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union, cast
import math
import enum
import warnings
import time

import pandas as pd
import torch
from torch import nn
from torch.utils.data import DataLoader
from torch import Tensor
import torch.nn.functional as F

from pytrial.data.patient_data import TabularPatientBase
from pytrial.utils.check import (
    check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist
)
from .base import TabularIndivBase, IndivTabDataset
from ..losses import XentLoss, BinaryXentLoss, MultilabelBinaryXentLoss, MSELoss

ModuleType = Union[str, Callable[..., nn.Module]]
_INTERNAL_ERROR_MESSAGE = 'Internal error. Please, open an issue.'

def _all_or_none(values):
    return all(x is None for x in values) or all(x is not None for x in values)

def _is_glu_activation(activation: ModuleType):
    return (
        isinstance(activation, str)
        and activation.endswith('GLU')
        or activation in [ReGLU, GEGLU]
    )

def reglu(x: Tensor) -> Tensor:
    """The ReGLU activation function from [1].
    References:
        [1] Noam Shazeer, "GLU Variants Improve Transformer", 2020
    """
    assert x.shape[-1] % 2 == 0
    a, b = x.chunk(2, dim=-1)
    return a * F.relu(b)


def geglu(x: Tensor) -> Tensor:
    """The GEGLU activation function from [1].
    References:
        [1] Noam Shazeer, "GLU Variants Improve Transformer", 2020
    """
    assert x.shape[-1] % 2 == 0
    a, b = x.chunk(2, dim=-1)
    return a * F.gelu(b)

class ReGLU(nn.Module):
    """The ReGLU activation function from [shazeer2020glu].
    Examples:
        .. testcode::
            module = ReGLU()
            x = torch.randn(3, 4)
            assert module(x).shape == (3, 2)
    References:
        * [shazeer2020glu] Noam Shazeer, "GLU Variants Improve Transformer", 2020
    """

    def forward(self, x: Tensor) -> Tensor:
        return reglu(x)


class GEGLU(nn.Module):
    """The GEGLU activation function from [shazeer2020glu].
    Examples:
        .. testcode::
            module = GEGLU()
            x = torch.randn(3, 4)
            assert module(x).shape == (3, 2)
    References:
        * [shazeer2020glu] Noam Shazeer, "GLU Variants Improve Transformer", 2020
    """

    def forward(self, x: Tensor) -> Tensor:
        return geglu(x)

def _make_nn_module(module_type: ModuleType, *args) -> nn.Module:
    if isinstance(module_type, str):
        if module_type == 'ReGLU':
            return ReGLU()
        elif module_type == 'GEGLU':
            return GEGLU()
        else:
            try:
                cls = getattr(nn, module_type)
            except AttributeError as err:
                raise ValueError(
                    f'Failed to construct the module {module_type} with the arguments {args}'
                ) from err
            return cls(*args)
    else:
        return module_type(*args)

class _TokenInitialization(enum.Enum):
    UNIFORM = 'uniform'
    NORMAL = 'normal'

    @classmethod
    def from_str(cls, initialization: str) -> '_TokenInitialization':
        try:
            return cls(initialization)
        except ValueError:
            valid_values = [x.value for x in _TokenInitialization]
            raise ValueError(f'initialization must be one of {valid_values}')

    def apply(self, x: Tensor, d: int) -> None:
        d_sqrt_inv = 1 / math.sqrt(d)
        if self == _TokenInitialization.UNIFORM:
            # used in the paper "Revisiting Deep Learning Models for Tabular Data";
            # is equivalent to `nn.init.kaiming_uniform_(x, a=math.sqrt(5))` (which is
            # used by torch to initialize nn.Linear.weight, for example)
            nn.init.uniform_(x, a=-d_sqrt_inv, b=d_sqrt_inv)
        elif self == _TokenInitialization.NORMAL:
            nn.init.normal_(x, std=d_sqrt_inv)

class CategoricalFeatureTokenizer(nn.Module):
    """Transforms categorical features to tokens (embeddings).
    See `FeatureTokenizer` for the illustration.
    The module efficiently implements a collection of `torch.nn.Embedding` (with
    optional biases).
    Examples:
        .. testcode::
            # the input must contain integers. For example, if the first feature can
            # take 3 distinct values, then its cardinality is 3 and the first column
            # must contain values from the range `[0, 1, 2]`.
            cardinalities = [3, 10]
            x = torch.tensor([
                [0, 5],
                [1, 7],
                [0, 2],
                [2, 4]
            ])
            n_objects, n_features = x.shape
            d_token = 3
            tokenizer = CategoricalFeatureTokenizer(cardinalities, d_token, True, 'uniform')
            tokens = tokenizer(x)
            assert tokens.shape == (n_objects, n_features, d_token)
    """

    category_offsets: Tensor

    def __init__(
        self,
        cardinalities: List[int],
        d_token: int,
        bias: bool,
        initialization: str,
    ) -> None:
        """
        Args:
            cardinalities: the number of distinct values for each feature. For example,
                :code:`cardinalities=[3, 4]` describes two features: the first one can
                take values in the range :code:`[0, 1, 2]` and the second one can take
                values in the range :code:`[0, 1, 2, 3]`.
            d_token: the size of one token.
            bias: if `True`, for each feature, a trainable vector is added to the
                embedding regardless of feature value. The bias vectors are not shared
                between features.
            initialization: initialization policy for parameters. Must be one of
                :code:`['uniform', 'normal']`. Let :code:`s = d ** -0.5`. Then, the
                corresponding distributions are :code:`Uniform(-s, s)` and :code:`Normal(0, s)`. In
                the paper [gorishniy2021revisiting], the 'uniform' initialization was
                used.
        References:
            * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko, "Revisiting Deep Learning Models for Tabular Data", 2021
        """
        super().__init__()
        assert cardinalities, 'cardinalities must be non-empty'
        assert d_token > 0, 'd_token must be positive'
        initialization_ = _TokenInitialization.from_str(initialization)

        category_offsets = torch.tensor([0] + cardinalities[:-1]).cumsum(0)
        self.register_buffer('category_offsets', category_offsets, persistent=False)
        self.embeddings = nn.Embedding(sum(cardinalities), d_token)
        self.bias = nn.Parameter(Tensor(len(cardinalities), d_token)) if bias else None

        for parameter in [self.embeddings.weight, self.bias]:
            if parameter is not None:
                initialization_.apply(parameter, d_token)

    @property
    def n_tokens(self) -> int:
        """The number of tokens."""
        return len(self.category_offsets)

    @property
    def d_token(self) -> int:
        """The size of one token."""
        return self.embeddings.embedding_dim

    def forward(self, x: Tensor) -> Tensor:
        x = self.embeddings(x + self.category_offsets[None])
        if self.bias is not None:
            x = x + self.bias[None]
        return x

class NumericalFeatureTokenizer(nn.Module):
    """Transforms continuous features to tokens (embeddings).
    See `FeatureTokenizer` for the illustration.
    For one feature, the transformation consists of two steps:
    * the feature is multiplied by a trainable vector
    * another trainable vector is added
    Note that each feature has its separate pair of trainable vectors, i.e. the vectors
    are not shared between features.
    Examples:
        .. testcode::
            x = torch.randn(4, 2)
            n_objects, n_features = x.shape
            d_token = 3
            tokenizer = NumericalFeatureTokenizer(n_features, d_token, True, 'uniform')
            tokens = tokenizer(x)
            assert tokens.shape == (n_objects, n_features, d_token)
    """

    def __init__(
        self,
        n_features: int,
        d_token: int,
        bias: bool,
        initialization: str,
    ) -> None:
        """
        Args:
            n_features: the number of continuous (scalar) features
            d_token: the size of one token
            bias: if `False`, then the transformation will include only multiplication.
                **Warning**: :code:`bias=False` leads to significantly worse results for
                Transformer-like (token-based) architectures.
            initialization: initialization policy for parameters. Must be one of
                :code:`['uniform', 'normal']`. Let :code:`s = d ** -0.5`. Then, the
                corresponding distributions are :code:`Uniform(-s, s)` and :code:`Normal(0, s)`.
                In [gorishniy2021revisiting], the 'uniform' initialization was used.
        References:
            * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko, "Revisiting Deep Learning Models for Tabular Data", 2021
        """
        super().__init__()
        initialization_ = _TokenInitialization.from_str(initialization)
        self.weight = nn.Parameter(Tensor(n_features, d_token))
        self.bias = nn.Parameter(Tensor(n_features, d_token)) if bias else None
        for parameter in [self.weight, self.bias]:
            if parameter is not None:
                initialization_.apply(parameter, d_token)

    @property
    def n_tokens(self) -> int:
        """The number of tokens."""
        return len(self.weight)

    @property
    def d_token(self) -> int:
        """The size of one token."""
        return self.weight.shape[1]

    def forward(self, x: Tensor) -> Tensor:
        x = self.weight[None] * x[..., None]
        if self.bias is not None:
            x = x + self.bias[None]
        return x

class FeatureTokenizer(nn.Module):
    def __init__(
        self,
        n_num_features: int,
        cat_cardinalities: List[int],
        d_token: int,
    ) -> None:
        """
        Args:
            n_num_features: the number of continuous features. Pass :code:`0` if there
                are no numerical features.
            cat_cardinalities: the number of unique values for each feature. See
                `CategoricalFeatureTokenizer` for details. Pass an empty list if there
                are no categorical features.
            d_token: the size of one token.
        """
        super().__init__()
        assert n_num_features >= 0, 'n_num_features must be non-negative'
        assert (
            n_num_features or cat_cardinalities
        ), 'at least one of n_num_features or cat_cardinalities must be positive/non-empty'
        self.initialization = 'uniform'
        self.num_tokenizer = (
            NumericalFeatureTokenizer(
                n_features=n_num_features,
                d_token=d_token,
                bias=True,
                initialization=self.initialization,
            )
            if n_num_features
            else None
        )
        self.cat_tokenizer = (
            CategoricalFeatureTokenizer(
                cat_cardinalities, d_token, True, self.initialization
            )
            if cat_cardinalities
            else None
        )

    @property
    def n_tokens(self) -> int:
        """The number of tokens."""
        return sum(
            x.n_tokens
            for x in [self.num_tokenizer, self.cat_tokenizer]
            if x is not None
        )

    @property
    def d_token(self) -> int:
        """The size of one token."""
        return (
            self.cat_tokenizer.d_token  # type: ignore
            if self.num_tokenizer is None
            else self.num_tokenizer.d_token
        )

    def forward(self, x_num: Optional[Tensor], x_cat: Optional[Tensor]) -> Tensor:
        """Perform the forward pass.
        Args:
            x_num: continuous features. Must be presented if :code:`n_num_features > 0`
                was passed to the constructor.
            x_cat: categorical features (see `CategoricalFeatureTokenizer.forward` for
                details). Must be presented if non-empty :code:`cat_cardinalities` was
                passed to the constructor.
        Returns:
            tokens
        Raises:
            AssertionError: if the described requirements for the inputs are not met.
        """
        assert (
            x_num is not None or x_cat is not None
        ), 'At least one of x_num and x_cat must be presented'
        assert _all_or_none(
            [self.num_tokenizer, x_num]
        ), 'If self.num_tokenizer is (not) None, then x_num must (not) be None'
        assert _all_or_none(
            [self.cat_tokenizer, x_cat]
        ), 'If self.cat_tokenizer is (not) None, then x_cat must (not) be None'
        x = []
        if self.num_tokenizer is not None:
            x.append(self.num_tokenizer(x_num))
        if self.cat_tokenizer is not None:
            x.append(self.cat_tokenizer(x_cat))
        return x[0] if len(x) == 1 else torch.cat(x, dim=1)


class CLSToken(nn.Module):
    """[CLS]-token for BERT-like inference.
    To learn about the [CLS]-based inference, see [devlin2018bert].
    When used as a module, the [CLS]-token is appended **to the end** of each item in
    the batch.
    Examples:
        .. testcode::
            batch_size = 2
            n_tokens = 3
            d_token = 4
            cls_token = CLSToken(d_token, 'uniform')
            x = torch.randn(batch_size, n_tokens, d_token)
            x = cls_token(x)
            assert x.shape == (batch_size, n_tokens + 1, d_token)
            assert (x[:, -1, :] == cls_token.expand(len(x))).all()
    References:
        * [devlin2018bert] Jacob Devlin, Ming-Wei Chang, Kenton Lee, Kristina Toutanova "BERT: Pre-training of Deep Bidirectional Transformers for Language Understanding" 2018
    """

    def __init__(self, d_token: int, initialization: str) -> None:
        """
        Args:
            d_token: the size of token
            initialization: initialization policy for parameters. Must be one of
                :code:`['uniform', 'normal']`. Let :code:`s = d ** -0.5`. Then, the
                corresponding distributions are :code:`Uniform(-s, s)` and :code:`Normal(0, s)`. In
                the paper [gorishniy2021revisiting], the 'uniform' initialization was
                used.
        References:
            * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko "Revisiting Deep Learning Models for Tabular Data", 2021
        """
        super().__init__()
        initialization_ = _TokenInitialization.from_str(initialization)
        self.weight = nn.Parameter(Tensor(d_token))
        initialization_.apply(self.weight, d_token)

    def expand(self, *leading_dimensions: int) -> Tensor:
        """Expand (repeat) the underlying [CLS]-token to a tensor with the given leading dimensions.
        A possible use case is building a batch of [CLS]-tokens. See `CLSToken` for
        examples of usage.
        Note:
            Under the hood, the `torch.Tensor.expand` method is applied to the
            underlying :code:`weight` parameter, so gradients will be propagated as
            expected.
        Args:
            leading_dimensions: the additional new dimensions
        Returns:
            tensor of the shape :code:`(*leading_dimensions, len(self.weight))`
        """
        if not leading_dimensions:
            return self.weight
        new_dims = (1,) * (len(leading_dimensions) - 1)
        return self.weight.view(*new_dims, -1).expand(*leading_dimensions, -1)

    def forward(self, x: Tensor) -> Tensor:
        """Append self **to the end** of each item in the batch (see `CLSToken`)."""
        return torch.cat([x, self.expand(len(x), 1)], dim=1)

class MultiheadAttention(nn.Module):
    """Multihead Attention (self-/cross-) with optional 'linear' attention.
    To learn more about Multihead Attention, see [devlin2018bert]. See the implementation
    of `Transformer` and the examples below to learn how to use the compression technique
    from [wang2020linformer] to speed up the module when the number of tokens is large.
    Examples:
        .. testcode::
            n_objects, n_tokens, d_token = 2, 3, 12
            n_heads = 6
            a = torch.randn(n_objects, n_tokens, d_token)
            b = torch.randn(n_objects, n_tokens * 2, d_token)
            module = MultiheadAttention(
                d_token=d_token, n_heads=n_heads, dropout=0.2, bias=True, initialization='kaiming'
            )
            # self-attention
            x, attention_stats = module(a, a, None, None)
            assert x.shape == a.shape
            assert attention_stats['attention_probs'].shape == (n_objects * n_heads, n_tokens, n_tokens)
            assert attention_stats['attention_logits'].shape == (n_objects * n_heads, n_tokens, n_tokens)
            # cross-attention
            assert module(a, b, None, None)
            # Linformer self-attention with the 'headwise' sharing policy
            k_compression = torch.nn.Linear(n_tokens, n_tokens // 4)
            v_compression = torch.nn.Linear(n_tokens, n_tokens // 4)
            assert module(a, a, k_compression, v_compression)
            # Linformer self-attention with the 'key-value' sharing policy
            kv_compression = torch.nn.Linear(n_tokens, n_tokens // 4)
            assert module(a, a, kv_compression, kv_compression)
    References:
        * [devlin2018bert] Jacob Devlin, Ming-Wei Chang, Kenton Lee, Kristina Toutanova "BERT: Pre-training of Deep Bidirectional Transformers for Language Understanding" 2018
        * [wang2020linformer] Sinong Wang, Belinda Z. Li, Madian Khabsa, Han Fang, Hao Ma "Linformer: Self-Attention with Linear Complexity", 2020
    """

    def __init__(
        self,
        *,
        d_token: int,
        n_heads: int,
        dropout: float,
        bias: bool,
        initialization: str,
    ) -> None:
        """
        Args:
            d_token: the token size. Must be a multiple of :code:`n_heads`.
            n_heads: the number of heads. If greater than 1, then the module will have
                an addition output layer (so called "mixing" layer).
            dropout: dropout rate for the attention map. The dropout is applied to
                *probabilities* and do not affect logits.
            bias: if `True`, then input (and output, if presented) layers also have bias.
                `True` is a reasonable default choice.
            initialization: initialization for input projection layers. Must be one of
                :code:`['kaiming', 'xavier']`. `kaiming` is a reasonable default choice.
        Raises:
            AssertionError: if requirements for the inputs are not met.
        """
        super().__init__()
        if n_heads > 1:
            assert d_token % n_heads == 0, 'd_token must be a multiple of n_heads'
        assert initialization in ['kaiming', 'xavier']

        self.W_q = nn.Linear(d_token, d_token, bias)
        self.W_k = nn.Linear(d_token, d_token, bias)
        self.W_v = nn.Linear(d_token, d_token, bias)
        self.W_out = nn.Linear(d_token, d_token, bias) if n_heads > 1 else None
        self.n_heads = n_heads
        self.dropout = nn.Dropout(dropout) if dropout else None

        for m in [self.W_q, self.W_k, self.W_v]:
            # the "xavier" branch tries to follow torch.nn.MultiheadAttention;
            # the second condition checks if W_v plays the role of W_out; the latter one
            # is initialized with Kaiming in torch
            if initialization == 'xavier' and (
                m is not self.W_v or self.W_out is not None
            ):
                # gain is needed since W_qkv is represented with 3 separate layers (it
                # implies different fan_out)
                nn.init.xavier_uniform_(m.weight, gain=1 / math.sqrt(2))
            if m.bias is not None:
                nn.init.zeros_(m.bias)
        if self.W_out is not None:
            nn.init.zeros_(self.W_out.bias)

    def _reshape(self, x: Tensor) -> Tensor:
        batch_size, n_tokens, d = x.shape
        d_head = d // self.n_heads
        return (
            x.reshape(batch_size, n_tokens, self.n_heads, d_head)
            .transpose(1, 2)
            .reshape(batch_size * self.n_heads, n_tokens, d_head)
        )

    def forward(
        self,
        x_q: Tensor,
        x_kv: Tensor,
        key_compression: Optional[nn.Linear],
        value_compression: Optional[nn.Linear],
    ) -> Tuple[Tensor, Dict[str, Tensor]]:
        """Perform the forward pass.
        Args:
            x_q: query tokens
            x_kv: key-value tokens
            key_compression: Linformer-style compression for keys
            value_compression: Linformer-style compression for values
        Returns:
            (tokens, attention_stats)
        """
        assert _all_or_none(
            [key_compression, value_compression]
        ), 'If key_compression is (not) None, then value_compression must (not) be None'
        q, k, v = self.W_q(x_q), self.W_k(x_kv), self.W_v(x_kv)
        for tensor in [q, k, v]:
            assert tensor.shape[-1] % self.n_heads == 0, _INTERNAL_ERROR_MESSAGE
        if key_compression is not None:
            k = key_compression(k.transpose(1, 2)).transpose(1, 2)
            v = value_compression(v.transpose(1, 2)).transpose(1, 2)  # type: ignore

        batch_size = len(q)
        d_head_key = k.shape[-1] // self.n_heads
        d_head_value = v.shape[-1] // self.n_heads
        n_q_tokens = q.shape[1]

        q = self._reshape(q)
        k = self._reshape(k)
        attention_logits = q @ k.transpose(1, 2) / math.sqrt(d_head_key)
        attention_probs = F.softmax(attention_logits, dim=-1)
        if self.dropout is not None:
            attention_probs = self.dropout(attention_probs)
        x = attention_probs @ self._reshape(v)
        x = (
            x.reshape(batch_size, self.n_heads, n_q_tokens, d_head_value)
            .transpose(1, 2)
            .reshape(batch_size, n_q_tokens, self.n_heads * d_head_value)
        )
        if self.W_out is not None:
            x = self.W_out(x)
        return x, {
            'attention_logits': attention_logits,
            'attention_probs': attention_probs,
        }

class Transformer(nn.Module):
    """Transformer with extra features.
    This module is the backbone of `FTTransformer`."""

    WARNINGS = {'first_prenormalization': True, 'prenormalization': True}

    class FFN(nn.Module):
        """The Feed-Forward Network module used in every `Transformer` block."""

        def __init__(
            self,
            *,
            d_token: int,
            d_hidden: int,
            bias_first: bool,
            bias_second: bool,
            dropout: float,
            activation: ModuleType,
        ):
            super().__init__()
            self.linear_first = nn.Linear(
                d_token,
                d_hidden * (2 if _is_glu_activation(activation) else 1),
                bias_first,
            )
            self.activation = _make_nn_module(activation)
            self.dropout = nn.Dropout(dropout)
            self.linear_second = nn.Linear(d_hidden, d_token, bias_second)

        def forward(self, x: Tensor) -> Tensor:
            x = self.linear_first(x)
            x = self.activation(x)
            x = self.dropout(x)
            x = self.linear_second(x)
            return x

    class Head(nn.Module):
        """The final module of the `Transformer` that performs BERT-like inference."""

        def __init__(
            self,
            *,
            d_in: int,
            bias: bool,
            activation: ModuleType,
            normalization: ModuleType,
            d_out: int,
        ):
            super().__init__()
            self.normalization = _make_nn_module(normalization, d_in)
            self.activation = _make_nn_module(activation)
            self.linear = nn.Linear(d_in, d_out, bias)

        def forward(self, x: Tensor) -> Tensor:
            x = x[:, -1]
            x = self.normalization(x)
            x = self.activation(x)
            x = self.linear(x)
            return x

    def __init__(
        self,
        *,
        d_token: int,
        n_blocks: int,
        attention_n_heads: int,
        attention_dropout: float,
        attention_initialization: str,
        attention_normalization: str,
        ffn_d_hidden: int,
        ffn_dropout: float,
        ffn_activation: str,
        ffn_normalization: str,
        residual_dropout: float,
        prenormalization: bool,
        first_prenormalization: bool,
        last_layer_query_idx: Union[None, List[int], slice],
        n_tokens: Optional[int],
        kv_compression_ratio: Optional[float],
        kv_compression_sharing: Optional[str],
        head_activation: ModuleType,
        head_normalization: ModuleType,
        d_out: int,
    ) -> None:
        super().__init__()
        if isinstance(last_layer_query_idx, int):
            raise ValueError(
                'last_layer_query_idx must be None, list[int] or slice. '
                f'Do you mean last_layer_query_idx=[{last_layer_query_idx}] ?'
            )
        if not prenormalization:
            assert (
                not first_prenormalization
            ), 'If `prenormalization` is False, then `first_prenormalization` must be False'
        assert _all_or_none([n_tokens, kv_compression_ratio, kv_compression_sharing]), (
            'If any of the following arguments is (not) None, then all of them must (not) be None: '
            'n_tokens, kv_compression_ratio, kv_compression_sharing'
        )
        assert kv_compression_sharing in [None, 'headwise', 'key-value', 'layerwise']
        if not prenormalization:
            if self.WARNINGS['prenormalization']:
                warnings.warn(
                    'prenormalization is set to False. Are you sure about this? '
                    'The training can become less stable. '
                    'You can turn off this warning by tweaking the '
                    'rtdl.Transformer.WARNINGS dictionary.',
                    UserWarning,
                )
            assert (
                not first_prenormalization
            ), 'If prenormalization is False, then first_prenormalization is ignored and must be set to False'
        if (
            prenormalization
            and first_prenormalization
            and self.WARNINGS['first_prenormalization']
        ):
            warnings.warn(
                'first_prenormalization is set to True. Are you sure about this? '
                'For example, the vanilla FTTransformer with '
                'first_prenormalization=True performs SIGNIFICANTLY worse. '
                'You can turn off this warning by tweaking the '
                'rtdl.Transformer.WARNINGS dictionary.',
                UserWarning,
            )
            time.sleep(3)

        def make_kv_compression():
            assert (
                n_tokens and kv_compression_ratio
            ), _INTERNAL_ERROR_MESSAGE  # for mypy
            # https://github.com/pytorch/fairseq/blob/1bba712622b8ae4efb3eb793a8a40da386fe11d0/examples/linformer/linformer_src/modules/multihead_linear_attention.py#L83
            return nn.Linear(n_tokens, int(n_tokens * kv_compression_ratio), bias=False)

        self.shared_kv_compression = (
            make_kv_compression()
            if kv_compression_ratio and kv_compression_sharing == 'layerwise'
            else None
        )

        self.prenormalization = prenormalization
        self.last_layer_query_idx = last_layer_query_idx

        self.blocks = nn.ModuleList([])
        for layer_idx in range(n_blocks):
            layer = nn.ModuleDict(
                {
                    'attention': MultiheadAttention(
                        d_token=d_token,
                        n_heads=attention_n_heads,
                        dropout=attention_dropout,
                        bias=True,
                        initialization=attention_initialization,
                    ),
                    'ffn': Transformer.FFN(
                        d_token=d_token,
                        d_hidden=ffn_d_hidden,
                        bias_first=True,
                        bias_second=True,
                        dropout=ffn_dropout,
                        activation=ffn_activation,
                    ),
                    'attention_residual_dropout': nn.Dropout(residual_dropout),
                    'ffn_residual_dropout': nn.Dropout(residual_dropout),
                    'output': nn.Identity(),  # for hooks-based introspection
                }
            )
            if layer_idx or not prenormalization or first_prenormalization:
                layer['attention_normalization'] = _make_nn_module(
                    attention_normalization, d_token
                )
            layer['ffn_normalization'] = _make_nn_module(ffn_normalization, d_token)
            if kv_compression_ratio and self.shared_kv_compression is None:
                layer['key_compression'] = make_kv_compression()
                if kv_compression_sharing == 'headwise':
                    layer['value_compression'] = make_kv_compression()
                else:
                    assert (
                        kv_compression_sharing == 'key-value'
                    ), _INTERNAL_ERROR_MESSAGE
            self.blocks.append(layer)

        self.head = Transformer.Head(
            d_in=d_token,
            d_out=d_out,
            bias=True,
            activation=head_activation,  # type: ignore
            normalization=head_normalization if prenormalization else 'Identity',
        )

    def _get_kv_compressions(self, layer):
        return (
            (self.shared_kv_compression, self.shared_kv_compression)
            if self.shared_kv_compression is not None
            else (layer['key_compression'], layer['value_compression'])
            if 'key_compression' in layer and 'value_compression' in layer
            else (layer['key_compression'], layer['key_compression'])
            if 'key_compression' in layer
            else (None, None)
        )

    def _start_residual(self, layer, stage, x):
        assert stage in ['attention', 'ffn'], _INTERNAL_ERROR_MESSAGE
        x_residual = x
        if self.prenormalization:
            norm_key = f'{stage}_normalization'
            if norm_key in layer:
                x_residual = layer[norm_key](x_residual)
        return x_residual

    def _end_residual(self, layer, stage, x, x_residual):
        assert stage in ['attention', 'ffn'], _INTERNAL_ERROR_MESSAGE
        x_residual = layer[f'{stage}_residual_dropout'](x_residual)
        x = x + x_residual
        if not self.prenormalization:
            x = layer[f'{stage}_normalization'](x)
        return x

    def forward(self, x: Tensor) -> Tensor:
        assert (
            x.ndim == 3
        ), 'The input must have 3 dimensions: (n_objects, n_tokens, d_token)'
        for layer_idx, layer in enumerate(self.blocks):
            layer = cast(nn.ModuleDict, layer)

            query_idx = (
                self.last_layer_query_idx if layer_idx + 1 == len(self.blocks) else None
            )
            x_residual = self._start_residual(layer, 'attention', x)
            x_residual, _ = layer['attention'](
                x_residual if query_idx is None else x_residual[:, query_idx],
                x_residual,
                *self._get_kv_compressions(layer),
            )
            if query_idx is not None:
                x = x[:, query_idx]
            x = self._end_residual(layer, 'attention', x, x_residual)

            x_residual = self._start_residual(layer, 'ffn', x)
            x_residual = layer['ffn'](x_residual)
            x = self._end_residual(layer, 'ffn', x, x_residual)
            x = layer['output'](x)

        x = self.head(x)
        return x


class BuildModel(nn.Module):
    def __init__(self,
        n_num_features,
        cat_cardinalities,
        output_dim,
        hidden_dim=128,
        num_layer=2,
        attention_dropout=0,
        attention_n_heads=8,
        attention_initialization='kaiming',
        attention_normalization='LayerNorm',
        ffn_dim=256,
        ffn_dropout=0,
        ffn_activation='ReGLU',
        ffn_normalization='LayerNorm',
        residual_dropout=0,
        prenormalization=True,
        first_prenormalization=False,
        last_layer_query_idx=None,
        n_tokens=None,
        kv_compression_ratio=None,
        kv_compression_sharing=None,
        head_activation='ReLU',
        head_normalization='LayerNorm',
        **kwargs,
        ) -> None:
        super().__init__()
        
        self.feature_tokenizer = FeatureTokenizer(
            n_num_features=n_num_features,
            cat_cardinalities=cat_cardinalities,
            d_token=hidden_dim,
        )
        self.cls_token = CLSToken(
            self.feature_tokenizer.d_token, self.feature_tokenizer.initialization
        )

        self.transformer = Transformer(
            d_token=hidden_dim,
            n_blocks=num_layer,
            attention_n_heads=attention_n_heads,
            attention_dropout=attention_dropout,
            attention_initialization=attention_initialization,
            attention_normalization=attention_normalization,
            ffn_d_hidden=ffn_dim,
            ffn_dropout=ffn_dropout,
            ffn_activation=ffn_activation,
            ffn_normalization=ffn_normalization,
            residual_dropout=residual_dropout,
            prenormalization=prenormalization,
            first_prenormalization=first_prenormalization,
            last_layer_query_idx=last_layer_query_idx,
            n_tokens=n_tokens,
            kv_compression_ratio=kv_compression_ratio,
            kv_compression_sharing=kv_compression_sharing,
            head_activation=head_activation,
            head_normalization=head_normalization,
            d_out=output_dim,
            )
    
    def forward(self, x_num, x_cat):
        # ensure correct data type for embedding
        x_cat = x_cat.long() 
        x_num = x_num.float()
        x = self.feature_tokenizer(x_num, x_cat)
        x = self.cls_token(x)
        x = self.transformer(x)
        return x

class FTTransformer(TabularIndivBase):
    '''
    Implement ft-transformer model for tabular individual outcome
    prediction in clinical trials [1]_. 

    Parameters
    ----------
    num_feat: list[str]
        the list of numerical feature names.

    cat_feat: list[str]
        the list of categorical feature names.

    cat_cardinalities: list[int]
        A list of categorical features' cardinalities.

    output_dim: int
        Dimension of the outputs. When doing classification, it equals to number of classes.

    mode: str
        The task's objectives, in `binary`, `multiclass`, `multilabel`, or `regression`

    hidden_dim: int
        Hidden dimensions of neural networks. Must be a multiple of :code:`n_heads=8`.

    num_layer: int
        Number of hidden layers.

    attention_dropout: float
        the dropout for attention blocks.
        Usually, positive values work better (even when the number of features is low).

    ffn_dim: int
        the *input* size for the *second* linear layer in `Transformer.FFN`.
        Note that it can be different from the output size of the first linear
        layer, since activations such as ReGLU or GEGLU change the size of input.
        For example, if :code:`ffn_d_hidden=10` and the activation is ReGLU (which
        is always true for the baseline and default configurations), then the
        output size of the first linear layer will be set to :code:`20`.

    ffn_dropout: float
        the dropout rate after the first linear layer in `Transformer.FFN`.

    residual_dropout: float
        the dropout rate for the output of each residual branch of all Transformer blocks.

    learning_rate: float
        Learning rate for optimization based on SGD. Use torch.optim.Adam by default.

    weight_decay: float
        Regularization strength for l2 norm; must be a positive float.
        Smaller values specify weaker regularization.
    
    batch_size: int
        Batch size when doing SGD optimization.

    epochs: int
        Maximum number of iterations taken for the solvers to converge.

    num_worker: int
        Number of workers used to do dataloading during training.

    device: str
        Target device to train the model, as `cuda:0` or `cpu`.

    experiment_id: str, optional (default='test')
        The name of current experiment. Decide the saved model checkpoint name.
    
    Notes
    -----
    .. [1] Gorishniy, Y., et al. (2021). Revisiting deep learning models for tabular data. NeurIPS'21.
    '''
    def __init__(self,
        num_feat,
        cat_feat,
        cat_cardinalities,
        output_dim,
        mode,
        hidden_dim=128,
        num_layer=2,
        attention_dropout=0,
        ffn_dim=256,
        ffn_dropout=0,
        residual_dropout=0,
        learning_rate=1e-4,
        weight_decay=1e-4,
        batch_size=64,
        epochs=10,
        num_worker=0,
        device='cuda:0',
        experiment_id='test'):
        super().__init__(experiment_id)

        mode = mode.lower()
        assert mode in ['binary', 'multiclass', 'regression', 'multilabel']

        self.config = {
            'num_feat':num_feat,
            'cat_feat':cat_feat,
            'n_num_features':len(num_feat),
            'cat_cardinalities':cat_cardinalities,
            'output_dim':output_dim,
            'hidden_dim':hidden_dim,
            'num_layer':num_layer,
            'attention_dropout':attention_dropout,
            'ffn_dim':ffn_dim,
            'ffn_dropout':ffn_dropout,
            'residual_dropout':residual_dropout,
            'learning_rate':learning_rate,
            'batch_size':batch_size,
            'weight_decay':weight_decay,
            'epochs':epochs,
            'num_worker':num_worker,
            'experiment_id':experiment_id,
            'model_name': 'MLP',
            'device':device,
            'mode':mode,
        }
        self._save_config(self.config)
        self.device = device

    def fit(self, train_data, valid_data=None):
        '''Train FT-Transformer model to predict patient outcome
        with tabular input data.

        Parameters
        ----------
        train_data: dict
            {
            'x': TabularPatientBase or pd.DataFrame,
            'y': pd.Series or np.ndarray
            }
            
            - 'x' contain all patient features; 
            - 'y' contain label for each row.

        valid_data: dict
            Same as `train_data`.
            Validation data during the training for early stopping.
        '''

        self._input_data_check(train_data)
        self._build_model()
        
        x_feat, y = self._parse_input_data(train_data)
        train_data={'x':x_feat, 'y':y}

        if valid_data is not None:
            x_feat_va, y_va = self._parse_input_data(valid_data)
            valid_data = {'x':x_feat_va, 'y':y_va}
                
        self._fit_model(train_data=train_data, valid_data=valid_data)
    
    def predict(self, test_data):
        '''
        Make prediction probability based on the learned model.

        Parameters
        ----------
        test_data: Dict or TabularPatientBase or pd.DataFrame or torch.Tensor
            {'x': TabularPatientBase or pd.DataFrame or torch.Tensor}
            
            'x' contain all patient features.

        Returns
        -------
        ypred: np.ndarray or torch.Tensor
            Prediction probability for each patient.
            
            - For binary classification, return shape (n, );
            - For multiclass classification, return shape (n, n_class).

        '''

        if isinstance(test_data, torch.Tensor):
            return self.model(test_data)
        
        x_feat, y = self._parse_input_data(test_data)
        test_data = {'x':x_feat, 'y': y}
        dataset = self._build_dataset(test_data)
        dataloader = DataLoader(dataset, 
            batch_size=self.config['batch_size'], 
            shuffle=False, 
            num_workers=self.config['num_worker'],
            pin_memory=True,
            )
        ypred = self._predict_on_dataloader(dataloader)

        return ypred

    def save_model(self, output_dir=None):
        '''
        Save the learned ft-transformer model to the disk.

        Parameters
        ----------
        output_dir: str or None
            The dir to save the learned model.
            If set None, will save model to `self.checkout_dir`.
        '''
        if output_dir is not None:
            make_dir_if_not_exist(output_dir)
        else:
            output_dir = self.checkout_dir

        self._save_config(self.config, output_dir=output_dir)
        self._save_checkpoint({'model':self.model}, output_dir=output_dir)

    def load_model(self, checkpoint):
        '''
        Load the model from the given checkpoint dir.

        Parameters
        ----------
        checkpoint: str
            The input dir that stores the pretrained model.

            - If a directory, the only checkpoint file `*.pth.tar` will be loaded.
            - If a filepath, will load from this file.
        '''
        checkpoint_filename = check_checkpoint_file(checkpoint)
        config_filename = check_model_config_file(checkpoint)
        state_dict = torch.load(checkpoint_filename)
        if config_filename is not None:
            config = self._load_config(config_filename)
            self.config.update(config)
        self.model = state_dict['model']

    def _build_model(self):
        self.model = BuildModel(
            **self.config,
            )
        self.model.to(self.config['device'])

    def _build_dataset(self, inputs):
        df = inputs['x']
        
        if 'y' in inputs: label = inputs['y']
        else: label = None

        return FTIndivTabDataset(
                df, 
                num_cols=self.config['num_feat'], 
                cat_cols=self.config['cat_feat'] , 
                label=label,
            )

    def _parse_input_data(self, inputs):
        if isinstance(inputs, dict):
            if isinstance(inputs['x'], TabularPatientBase):
                dataset = inputs['x']
                x_feat = dataset.df
                y = inputs['y']
            else:
                x_feat = inputs['x']
                y = inputs['y']                
            
            if isinstance(y, pd.DataFrame) or isinstance(y, pd.Series):
                y = y.values
        
        if isinstance(inputs, pd.DataFrame) or isinstance(inputs, torch.Tensor):
            x_feat, y = inputs, None
        
        if isinstance(inputs, TabularPatientBase):
            x_feat, y = inputs.df, None

        return x_feat, y

    def _build_loss_model(self):
        mode = self.config['mode']
        if mode == 'multiclass':
            return FTXentLoss(self.model)

        elif mode == 'binary':
            return FTBinaryXentLoss(self.model)

        elif mode == 'regression':
            return FTMSELoss(self.model)

        elif mode == 'multilabe':
            return FTMultilabelBinaryXentLoss(self.model)

        else:
            raise ValueError(f'Do not recognize mode `{mode}`, please correct.')

    def _predict_on_dataloader(self, dataloader):
        pred_list, label_list = [], []
        for batch in dataloader:
            x_cat, x_num = batch['x_cat'], batch['x_num']
            x_num = x_num.to(self.device)
            x_cat = x_cat.to(self.device)
            if 'y' in batch: label_list.append(batch.pop('y'))
            pred = self.model(x_num, x_cat)
            pred_list.append(pred)
        pred = torch.cat(pred_list)
        label = torch.cat(label_list) if len(label_list) > 0 else None
        return {'pred':pred, 'label':label}


class FTIndivTabDataset(IndivTabDataset):
    def __init__(self, df, cat_cols, num_cols, label=None):
        self.df = df
        self.label = label
        self.cat_cols = cat_cols
        self.num_cols = num_cols
    
    def __len__(self):
        return len(self.df)
    
    def __getitem__(self, idx):
        if self.label is not None: 
            return {
                'x_cat': self.df.iloc[idx][self.cat_cols].values, 
                'x_num': self.df.iloc[idx][self.num_cols].values,
                'y': self.label[idx]
                }
        else: 
            return {
                'x_cat': self.df.iloc[idx][self.cat_cols].values, 
                'x_num': self.df.iloc[idx][self.num_cols].values
                }

'''
Customized loss models for FT-Transformer.
'''

class FTXentLoss(XentLoss):
    '''
    The basic xent loss
    for multi-class classification.
    '''
    def forward(self, inputs):
        logits = self.model(inputs['x_num'], inputs['x_cat'])
        loss = self.xent_loss(logits, inputs['y'])
        return {'loss_value':loss}

class FTBinaryXentLoss(BinaryXentLoss):
    '''
    The basic binary xent loss
    for binary classification.
    '''
    def forward(self, inputs):
        logits = self.model(inputs['x_num'], inputs['x_cat'])
        y = inputs['y']
        if y.shape != logits.shape:
            logits = logits.reshape(y.shape)
        loss = self.xent_loss(logits, y.float())
        return {'loss_value':loss}

class FTMultilabelBinaryXentLoss(MultilabelBinaryXentLoss):
    '''
    The basic binary xent loss
    for multilabel classification.
    '''
    def forward(self, inputs):
        logits = self.model(inputs['x_num'], inputs['x_cat'])
        loss = self.xent_loss(logits, inputs['y'].float())
        return {'loss_value':loss}


class FTMSELoss(MSELoss):
    def forward(self, inputs):
        logits = self.model(inputs['x_num'], inputs['x_cat'])
        loss = self.loss(logits, inputs['y'])
        return {'loss_value':loss}


from collections import defaultdict
import pdb

import numpy as np
from torch.utils.data import Dataset
import torch
from torch.nn.utils.rnn import pad_sequence

from .data_utils import collect_trials_from_topic

class Trial:
    '''
    Contains information about a single trial.

    A trial is associated with multiple properties.
    
    Parameters
    ----------
    nctid : str
        The unique identifier of a trial. Usually be the nctid linked to clnicaltrials.gov.

    title: str
        The title of the clinical trial.

    label: int
        The label of the clinical trial. 1 for success, 0 for failure. 

    status: str
        The status of the trial. Usually be 'Completed', 'Terminated', 'Withdrawn', 'Enrolling by invitation', 'Active, not recruiting', 'Recruiting', 'Suspended', 'Approved for marketing', 'Temporarily not available', 'Available', 'No longer available', 'Unknown status'.

    year: int
        The year when the trial starts.
    
    end_year: int (default=None)
        The year when the trial ends.

    phase: str (default=None)
        The phase of the trial. Usually be 'Phase 1', 'Phase 2', 'Phase 3', 'Phase 4', 'N/A'.

    diseases: list[str] (default=None)
        A list of diseases (in ICD codes) associated with the trial.
    
    drugs: list[str] (default=None)
        A list of drug names associated with the trial.
    
    smiles: list[str] (default=None)
        A list of SMILE string described the associated the drugs.

    inc_criteria: str (default=None)
        The inclusion criteria of the trial.
    
    exc_criteria: str (default=None)
        The exclusion criteria of the trial.

    description: str (default=None)
        The description of the trial.

    why_stop: str (default=None)
        The reason why a trial stops.

    Attributes
    ----------
    attr_dict: dict
        A dictionary of all the attributes of the trial.
    '''
    def __init__(self, nctid=None, title=None, label=None, status=None, year=None, end_year=None, phase=None, diseases=None, drugs=None, smiles=None, inc_criteria=None, exc_criteria=None, description=None, why_stop=None, **kwargs):
        self.attr_dict = {
            'nctid': nctid,
            'title': title,
            'label': label,
            'status': status,
            'year': year,
            'end_year': end_year,
            'phase': phase,
            'diseases': diseases,
            'drugs': drugs,
            'smiles': smiles,
            'inclusion_criteria': inc_criteria,
            'exclusion_criteria': exc_criteria,
            'description': description,
            'why_stop': why_stop,
        }
        self.attr_dict.update(kwargs)
    
    def __repr__(self):
        line = f"\n\tTrial with id {self.attr_dict['nctid']}:" \
            f"\n\ttitle: {self.attr_dict['title']}" \
                f"\n\tyear: {self.attr_dict['year']}" \
                    f"\n\tstatus: {self.attr_dict['status']}"
        return line

class Topic:
    '''
    Contains the information about a sequence of trial lists.

    A `Topic` is associated with multiple `Trial` which are conducted at different timestamps.

    Parameters
    ----------
    topic_id: str
        The topic this sequence of trial belongs to.
    
    trials: List[List[Trial]]
        A list of lists of trials. Each list of trials is conducted at the same timestamp.
    
    Attributes
    ----------
    topic_id: str
        The topic this sequence of trial belongs to.
    
    trials: List[List[Trial]]
        A list of lists of trials. Each list of trials is conducted at the same timestamp.
    '''
    def __init__(self, topic_id, trials):
        self.topic_id = topic_id
        self.trials = trials
        self.number_of_trials = sum([len(trial_step) for trial_step in self.trials])
    
    def __repr__(self):
        line = f"Trial topic id {self.topic_id}:" \
            f"\n\t# of timesteps: {len(self.trials)}; # of trials: {self.number_of_trials}."
        return line
    
    def __getitem__(self, idx):
        '''
        Get the trials at the given index of timestep.
        '''
        return self.trials[idx]
    
    def __len__(self):
        '''
        Get the number of timesteps.
        '''
        return len(self.trials)
    
    def add(self, topic):
        '''
        Add trials in a topic object to the current topic object.

        Parameters
        ----------
        topic: Topic
            The topic object to be added.
        '''
        assert self.topic_id == topic.topic_id, 'The topic id of the two topic objects are not the same.'
        i = 0
        for trial_step in topic.trials:
            timestamp = trial_step[0].attr_dict['year']
            while i < len(self.trials) and self.trials[i][0].attr_dict['year'] < timestamp:
                i += 1
            if i == len(self.trials):
                self.trials.append(trial_step) # append to the last of the list
            else:
                # insert into the specific timestamp, the year might not be the same as the timestamp
                # because added trials might be between the two timestamps in the original topic
                self.trials[i].extend(trial_step)     
        self.number_of_trials = sum([len(trial_step) for trial_step in self.trials])

class SequenceTrial(Dataset):
    '''
    A dataset consisting of multiple topics of clinical trials. Each topic is a list of clinical trials under this topic ordered with their timestamps.

    Parameters
    ----------
    topics: List[Topic]
        A list of `Topic` objects.
    
    Attributes
    ----------
    topics: List[Topic]
        A list of `Topic` objects.
    
    new_trials: List[Trial]
        A list of new trials that are not in the original topics after added.
        Every add operation will add an new list of trials to this list.
        This list is used to keep track of the new trials added to the dataset then
        can be used to evaluate the model's performance on newly added trials.
    '''
    def __init__(self, topics):
        self.topics = topics
        self.indices = np.array([topic.topic_id for topic in self.topics])
        self.new_trials = []

        # build trials from topics
        self.trials = defaultdict(list) # a dictionary of trials with topic id as key
        for topic in self.topics:
            nctids = collect_trials_from_topic(topic)
            self.trials[topic.topic_id].extend(nctids)
    
    def __len__(self):
        return len(self.topics)
    
    def __getitem__(self, idx):
        return self.topics[idx]

    def __repr__(self):
        number_of_trials = sum([topic.number_of_trials for topic in self.topics])
        line = f"TrialSequence: # of topics: {len(self.topics)}; # of trials: {number_of_trials}."
        return line

    def add(self, topics):
        '''
        Add a new topic or merge a topic of trials to the dataset.

        Parameters
        ----------
        topic: Topic or List[Topic]
            A `Topic` or list of `Topic` objects.
        '''
        if isinstance(topics, Topic):
            topics = [topics]
        
        new_trials = []
        for topic in topics:
            if topic.topic_id in self.indices:
                idx = np.where(self.indices == topic.topic_id)[0][0]
                self.topics[idx].add(topic)
            else:
                print(f'{topic.topic_id} not found, add a new topic.')
                self.topics.append(topic)
                self.indices = np.append(self.indices, topic.topic_id)
            trial_ids = collect_trials_from_topic(topic)
            new_trials.extend(trial_ids)
            self.trials[topic.topic_id].extend(trial_ids)
        self.new_trials.append(new_trials)


from typing import Any, Dict, List, Optional, Tuple, Union

from dataclasses import dataclass
import torch
from transformers.file_utils import ModelOutput

@dataclass
class EHRBartOutput(ModelOutput):
    """
    Base class for sequence-to-sequence language models outputs.

    Args:
        loss (:obj:`torch.FloatTensor` of shape :obj:`(1,)`, `optional`, returned when :obj:`labels` is provided):
            Language modeling loss.
        logits (:obj:`torch.FloatTensor` of shape :obj:`(batch_size, sequence_length, config.vocab_size)`):
            Prediction scores of the language modeling head (scores for each vocabulary token before SoftMax).
        past_key_values (:obj:`tuple(tuple(torch.FloatTensor))`, `optional`, returned when ``use_cache=True`` is passed or when ``config.use_cache=True``):
            Tuple of :obj:`tuple(torch.FloatTensor)` of length :obj:`config.n_layers`, with each tuple having 2 tensors
            of shape :obj:`(batch_size, num_heads, sequence_length, embed_size_per_head)`) and 2 additional tensors of
            shape :obj:`(batch_size, num_heads, encoder_sequence_length, embed_size_per_head)`.

            Contains pre-computed hidden-states (key and values in the self-attention blocks and in the cross-attention
            blocks) that can be used (see :obj:`past_key_values` input) to speed up sequential decoding.
        decoder_hidden_states (:obj:`tuple(torch.FloatTensor)`, `optional`, returned when ``output_hidden_states=True`` is passed or when ``config.output_hidden_states=True``):
            Tuple of :obj:`torch.FloatTensor` (one for the output of the embeddings + one for the output of each layer)
            of shape :obj:`(batch_size, sequence_length, hidden_size)`.

            Hidden-states of the decoder at the output of each layer plus the initial embedding outputs.
        decoder_attentions (:obj:`tuple(torch.FloatTensor)`, `optional`, returned when ``output_attentions=True`` is passed or when ``config.output_attentions=True``):
            Tuple of :obj:`torch.FloatTensor` (one for each layer) of shape :obj:`(batch_size, num_heads,
            sequence_length, sequence_length)`.

            Attentions weights of the decoder, after the attention softmax, used to compute the weighted average in the
            self-attention heads.
        cross_attentions (:obj:`tuple(torch.FloatTensor)`, `optional`, returned when ``output_attentions=True`` is passed or when ``config.output_attentions=True``):
            Tuple of :obj:`torch.FloatTensor` (one for each layer) of shape :obj:`(batch_size, num_heads,
            sequence_length, sequence_length)`.

            Attentions weights of the decoder's cross-attention layer, after the attention softmax, used to compute the
            weighted average in the cross-attention heads.
        encoder_last_hidden_state (:obj:`torch.FloatTensor` of shape :obj:`(batch_size, sequence_length, hidden_size)`, `optional`):
            Sequence of hidden-states at the output of the last layer of the encoder of the model.
        encoder_hidden_states (:obj:`tuple(torch.FloatTensor)`, `optional`, returned when ``output_hidden_states=True`` is passed or when ``config.output_hidden_states=True``):
            Tuple of :obj:`torch.FloatTensor` (one for the output of the embeddings + one for the output of each layer)
            of shape :obj:`(batch_size, sequence_length, hidden_size)`.

            Hidden-states of the encoder at the output of each layer plus the initial embedding outputs.
        encoder_attentions (:obj:`tuple(torch.FloatTensor)`, `optional`, returned when ``output_attentions=True`` is passed or when ``config.output_attentions=True``):
            Tuple of :obj:`torch.FloatTensor` (one for each layer) of shape :obj:`(batch_size, num_heads,
            sequence_length, sequence_length)`.

            Attentions weights of the encoder, after the attention softmax, used to compute the weighted average in the
            self-attention heads.
        perplexity:
            perplexity calculated when the label mask is given.
    """

    loss: Optional[torch.FloatTensor] = None
    logits: torch.FloatTensor = None
    past_key_values: Optional[Tuple[Tuple[torch.FloatTensor]]] = None
    decoder_hidden_states: Optional[Tuple[torch.FloatTensor]] = None
    decoder_attentions: Optional[Tuple[torch.FloatTensor]] = None
    cross_attentions: Optional[Tuple[torch.FloatTensor]] = None
    encoder_last_hidden_state: Optional[torch.FloatTensor] = None
    encoder_hidden_states: Optional[Tuple[torch.FloatTensor]] = None
    encoder_attentions: Optional[Tuple[torch.FloatTensor]] = None
    perplexity: Optional[torch.FloatTensor] = None



'''
Provide data functions for trial data simulation.
'''
from pytrial.data.patient_data import TabularPatientBase, SequencePatientBase

class Voc(object):
    def __init__(self):
        self.idx2word = {}
        self.word2idx = {}
    
    def __len__(self):
        return len(self.idx2word.keys())

    def add_sentence(self, sentence):
        for word in sentence:
            if word not in self.word2idx:
                self.idx2word[len(self.word2idx)] = word
                self.word2idx[word] = len(self.word2idx)

class TabularPatient(TabularPatientBase):
    '''
    Tabular patient records.

    IGNORE:
    # MedGAN etc.: only support discrete events, should map continous values to quantiles then do simulation.
    # CopulaGaussian, CTGAN etc.: support both discrete and continuos values.
    IGNORE
    '''
    def __init__(self, df, metadata=None, transform=True):
        super().__init__(df, metadata=metadata, transform=transform)


class SequencePatient(SequencePatientBase):
    '''
    Load sequential patient inputs for longitudinal patient records generation.

    Parameters
    ----------
    data: dict
        A dict contains patient data in sequence and/or in tabular.
        Given dict:
            {
                'x': np.ndarray or pd.DataFrame
                    Static patient features in tabular form, typically those baseline information.

                'v': list or np.ndarray
                    Patient visit sequence in dense format or in tensor format (depends on the model input requirement.)
                    If in dense format, it is like [[c1,c2,c3],[c4,c5],...], with shape [n_patient, NA, NA];
                    If in tensor format, it is like [[0,1,1],[1,1,0],...] (multi-hot encoded), 
                        with shape [n_patient, max_num_visit, max_num_event].
                
                'y': np.ndarray or pd.Series
                    Target label for each patient if making risk detection, with shape [n_patient, n_class];
                    Target label for each visit if making some visit-level prediction with shape [n_patient, NA, n_class].
            }
    
    metadata: dict (optional)
        A dict contains configuration of input patient data.
        metadata:
            {
                'voc': dict[Voc]
                    Vocabulary contains the event index to the exact event name, has three keys in general:
                    'diag', 'med', 'prod', corresponding to diagnosis, medication, and procedure.
                    ``Voc`` object should have two functions: `idx2word` and `word2idx`.
                
                'visit': dict[str]
                    a dict contains the format of input data for processing input visit sequences.
                    `visit`: {
                        'mode': 'tensor' or 'dense',
                        'order': list[str] (required when `mode='tensor'`)
                    },

                'label': dict[str]
                    a dict contains the format of input data for processing input labels.
                    `label`: {
                        'mode': 'tensor' or 'dense',
                    }
                
                'max_visit': int
                    the maximum number of visits considered when building tensor inputs, ignored
                    when visit mode is dense.
                
                'n_num_feature': int
                    the number of numerical features in patients' baseline features
                
                'cat_cardinalities': list[int]
                    the cardinalities of each categorical features
            }

    '''
    visit = None
    feature = None
    label = None
    max_visit = None
    visit_voc_size = None
    visit_order = None

    metadata = {
        'voc': {},
        
        'visit':{
            'mode': 'dense',
            'order': ['diag', 'prod', 'med'],
            },

        'label':{
            'mode': 'tensor',
            },

        'max_visit': 20,
    }

    def __init__(self, data, metadata=None) -> None:
        super().__init__(data=data, metadata=metadata)


import pdb

import torch
from torch import nn

class InfoNCELoss(nn.Module):
    '''
    The basic InfoNCE loss.
    Can be subclass and get other InfoNCE-based losses.
    '''

    def __init__(self, model, logit_scale_init_value) -> None:
        super().__init__()
        self.model = model
        self.logit_scale = nn.Parameter(torch.log(torch.tensor(1/logit_scale_init_value)))
        self.xent_loss = nn.CrossEntropyLoss()

    def forward(self, inputs):
        outputs = self.model(inputs)
        embs = outputs['embs']
        logits = self.compute_logits(embs, embs)
        loss = (self.compute_loss(logits) + self.compute_loss(logits.t())) / 2
        return {'loss_value':loss}

    def compute_loss(self, logits):
        return self.xent_loss(logits, torch.arange(len(logits), device=logits.device))

    def compute_logits(self, embs, target_embs):
        self.logit_scale.data = torch.clamp(self.logit_scale.data, 0, 4.6052)
        logit_scale = self.logit_scale.exp()
        logits = torch.matmul(self._normalize(embs), self._normalize(target_embs).t())
        logits = logits * logit_scale
        return logits

    def _normalize(self, emb):
        return emb / emb.norm(dim=-1, keepdim=True)


import abc
import pdb
import os
import json
import math
from collections import defaultdict

import numpy as np
import pandas as pd
import torch
from torch import nn
from torch.nn.utils.rnn import pad_sequence
from tqdm import tqdm

from pytrial.utils.check import check_model_dir
from pytrial.utils.check import make_dir_if_not_exist
from pytrial.tasks.trial_simulation.data import SequencePatient

class SequenceSimulationBase(abc.ABC):
    '''Abstract class for all sequential patient data simulations.
    
    Parameters
    ----------
    experiment_id: str, optional (default = 'test')
        The name of current experiment.
    '''
    training=False
    @abc.abstractmethod
    def __init__(self, experiment_id='test'):
        check_model_dir(experiment_id)
        self.checkout_dir = os.path.join('./experiments_records', experiment_id,
                                         'checkpoints')
        self.result_dir = os.path.join('./experiments_records', experiment_id,
                                       'results')
        make_dir_if_not_exist(self.checkout_dir)
        make_dir_if_not_exist(self.result_dir)
    
    @abc.abstractmethod
    def fit(self, train_data):
        '''Fit the model to the given training data. Need to be implemented by the child class.

        Parameters
        ----------
        train_data: SequencePatient
            The training data.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def predict(self, number_of_predictions):
        '''Generate synthetic sequence patient data. Need to be implemented by the child class.
        
        Parameters
        ----------
        number_of_predictions: int
            The number of synthetic sequence patient data to generate.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def save_model(self, output_dir):
        '''Save the model to the given directory. Need to be implemented by the child class.

        Parameters
        ----------
        output_dir: str
            The given directory to save the model.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def load_model(self, checkpoint):
        '''Load the model from the given directory. Need to be implemented by the child class.
        
        Parameters
        ----------
        checkpoint: str
            The given directory to load the model.
        '''
        raise NotImplementedError
    
    def train(self, mode=True):
        '''Set the model in training mode. Work similar to `model.train()` in PyTorch.
        
        Parameters
        ----------
        mode: bool, optional (default = True)
            Whether to set the model in training mode.
            ``False`` means the model is in evaluation mode.
            ``True`` means the model is in training mode.
        '''
        self.training = mode
        self.model.train()
        return self
    
    def eval(self, mode=False):
        '''Set the model in evaluation mode. Work similar to `model.eval()` in PyTorch.

        Parameters
        ----------
        mode: bool, optional (default = False)
            Whether to set the model in evaluation mode.
            ``False`` means the model is in evaluation mode.
            ``True`` means the model is in training mode.
        '''
        self.training = mode
        self.model.eval()
        return self

    def _input_data_check(self, inputs):
        assert isinstance(inputs, SequencePatient), 'Wrong input type.'
    
    def _save_checkpoint(self, state,
                        epoch_id=0,
                        is_best=False,
                        output_dir=None,
                        filename='checkpoint.pth.tar'):
        if output_dir is None:
            output_dir = self.checkout_dir

        if epoch_id < 1:
            filepath = os.path.join(output_dir, 'latest.' + filename)
        elif is_best:
            filepath = os.path.join(output_dir, 'best.' + filename)
        else:
            filepath = os.path.join(self.checkout_dir,
                                    str(epoch_id) + '.' + filename)
        torch.save(state, filepath)

    def _save_config(self, config, output_dir=None):
        if output_dir is None:
            output_dir = self.checkout_dir

        temp_path = os.path.join(output_dir, 'config.json')
        if os.path.exists(temp_path):
            os.remove(temp_path)
        with open(temp_path, 'w', encoding='utf-8') as f:
            f.write(
                json.dumps(config, indent=4)
            )

    def _load_config(self, checkpoint=None):
        '''
        Load model config from the given directory.
        Parameters
        ----------
        checkpoint: str
            The given filepath (e.g., ./checkpoint/config.json)
            to load the model config.
        '''
        if checkpoint is None:
            temp_path = os.path.join(checkpoint, 'config.json')
        else:
            temp_path = checkpoint

        assert os.path.exists(temp_path), 'Cannot find `config.json` under {}'.format(self.checkout_dir)
        with open(temp_path, 'r') as f:
            config = json.load(f)
        return config
    
    def _compute_n_per_sample(self, n_test_sample, n=None, n_per_sample=None):
        if n_per_sample is not None:
            n_total = n_test_sample*n_per_sample
            if n is not None:
                n_total = min(n_total, n)
            return n_total, n_per_sample
        else:
            return n, math.ceil(n_test_sample / n)

    def _get_num_visit(self, data, idx):
        visits = data['v']
        num_visit_list = []
        for k,v in visits.items():
            num_visit_list.append(len(v[idx]))
        
        num_visit_uq = list(set(num_visit_list))
        assert len(num_visit_uq) == 1, f'Find mismatch in the number of visit events {num_visit_list}, please check the input data {visits}.'
        return num_visit_uq[0]

    def _prepare_input(self, data, idx, vdx):
        '''
        Prepare inputs for sequence simulation models.

        Parameters
        ----------
        data: dict[list]
            A single patient records.

        idx: int
            The patient index.

        vdx: int
            The target visit index.

        '''
        visits = data['v']
        feature = data['x']
        if not isinstance(feature, torch.Tensor): feature = torch.tensor(feature)

        feature = feature.to(self.device)

        inputs = {
            'v':{},
            'y':{},
            'x':feature[idx], # baseline feature
            }

        for k in visits.keys():
            v = pad_sequence(list(map(torch.tensor,visits[k][idx][:vdx])), 
                    batch_first=True)
            
            y = torch.tensor(visits[k][idx][vdx])

            v = v.to(self.device)
            y = y.to(self.device)

            inputs['v'][k] = v
            inputs['y'][k] = y

        return inputs

class InputEventEmbedding(nn.Module):
    def __init__(self, orders, vocab_size, emb_size, padding_idx) -> None:
        super().__init__()
        # build input embeddings
        emb_dict = {}
        for i, order in enumerate(orders):
            emb_dict[order] = nn.Embedding(vocab_size[i], embedding_dim=emb_size, padding_idx=padding_idx)
        self.embeddings = nn.ModuleDict(emb_dict)
    
    def forward(self, inputs):
        emb_list = []
        for k, v in inputs['v'].items():
            emb = self.embeddings[k](v)
            emb_list.append(emb)
        embs = torch.cat(emb_list, 1)
        return embs

class GAN(nn.Module):
    '''
    Build generative adversarial network module referring to MedGAN.
    '''
    def __init__(self,
        emb_size,
        total_vocab_size,
        gen_dims,
        dis_dims,
        ) -> None:
        super().__init__()

        # generator and discriminator part
        self.gen_module = nn.ModuleDict({
            'fc1': nn.Linear(emb_size*2, gen_dims[0]), # from z+s -> hidden feat
            'fc2': nn.Linear(gen_dims[0], gen_dims[1]),
            'fc3': nn.Linear(gen_dims[1], total_vocab_size),
        })

        self.dis_module = nn.ModuleDict({
            'fc1': nn.Linear(total_vocab_size, dis_dims[0]), # x,\bar{x} -> {0,1}
            'fc2': nn.Linear(dis_dims[0], 1),
        })

    def forward(self, s):
        '''forward makes generation taking the state embeddings as inputs.
        '''
        z_random = torch.randn(s.size()).to(s.device)
        return self.infer_generator(s, z_random)
        
    def infer_generator(self, s, z):
        '''do generator inference, x
        args:
            s: state vector [bs, emb_size]
            z: visit or random [bs, emb_size]
        '''
        inputs = torch.cat([z,s], -1) # bs, emb_size*2
        tempVec = inputs
        h = self.gen_module['fc1'](tempVec) # bs, 256
        h = nn.ReLU()(h)

        tempVec = h + tempVec

        h = self.gen_module['fc2'](tempVec)
        h = nn.ReLU()(h)
        tempVec = h + tempVec # bs, 256

        h = self.gen_module['fc3'](tempVec)
        outputs = nn.Sigmoid()(h)
        return outputs

    def infer_discriminator(self, x):
        '''do discriminator inference
        args:
            x: the decoded fake records or the true records x
        '''
        h = self.dis_module['fc1'](x)
        h = nn.ReLU()(h)
        h = self.dis_module['fc2'](h)
        pred = h.sigmoid().squeeze(1)
        return pred


class RNN(nn.Module):
    RNN_TYPE = {
        'rnn':nn.RNN,
        'lstm':nn.LSTM,
        'gru':nn.GRU,
    }
    def __init__(self, 
        rnn_type,
        emb_size,
        num_layer,
        bidirectional,
        ) -> None:
        super().__init__()
        self.model = self.RNN_TYPE[rnn_type](
            input_size=emb_size,
            hidden_size=emb_size,
            num_layers=num_layer,
            bidirectional=bidirectional,
            batch_first=True,
        )
        self.bidirectional = bidirectional
    
    def forward(self, x):
        outputs = self.model(x)[0]
        return outputs


def transform_sequence_to_table(data, order, voc):
    # init the outputs
    outputs = defaultdict(list)
    outputs['pid'] = []
    outputs['vid'] = []
    columns = []
    for od in order:
        for k in voc[od].word2idx.keys():
            columns.append(f'{od}_{k}')

    output_list = []
    for pid, sample in tqdm(enumerate(data),desc='Transforming sequence to tabular format', total=len(data)):
        for vid, visit in enumerate(sample):
            visit_ = []
            for i, event_type in enumerate(order):
                visit_.extend([columns.index(f'{event_type}_{voc[event_type].idx2word[v]}') for v in visit[i]])

            visit_mh = np.zeros(len(columns))
            visit_mh[visit_] = 1
            outputs['pid'].append(pid)
            outputs['vid'].append(vid)
            output_list.append(visit_mh)

    output_df = pd.DataFrame(output_list, columns=columns)
    index_df = pd.DataFrame(outputs)
    output_df = pd.concat([index_df, output_df], axis=1)
    return output_df

def transform_table_to_sequence(data, order, voc):
    '''
    data: pd.DataFrame
    one column for each event type
    one row for each visit
    pid: patient id
    vid: visit id
    transform the outputs to the original format
    v = [[event1, event2, ...], [event1, event2, ...], ], [event1, event2, ...], ...]
    each v[i] is a patient
    each v[i][j] is a visit
    each v[i][j][k] is a list of that type of events
    '''
    outputs = []
    pid_list = []
    for pid, sample in tqdm(data.groupby('pid'), desc='Transforming tabular format to sequence', total=len(data['pid'].unique())):
        output = []
        for vid, visit in sample.groupby('vid'):
            visit_ = []
            for i, event_type in enumerate(order):
                visit_.append([voc[event_type].word2idx[k.split('_')[-1]] for k in visit.columns if k.startswith(event_type) and visit[k].values[0]==1])
            output.append(visit_)
        outputs.append(output)
        pid_list.append(pid)
    return outputs, pid_list



'''

(I). Trial_Dataset for prediction
(II). Trial_Dataset_Complete for interpretation
(III). SMILES lst 
(IV). disease lst icd-code 

'''

import torch, csv, os
from torch.utils import data 
from torch.utils.data.dataloader import default_collate
# from HINT.molecule_encode import smiles2mpnnfeature
from .protocol_encode import protocol2feature, load_sentence_2_vec

# sentence2vec = load_sentence_2_vec() 
sentence2vec = dict() 

class Trial_Dataset(data.Dataset):
	def __init__(self, nctid_lst, label_lst, smiles_lst, icdcode_lst, criteria_lst):
		self.nctid_lst = nctid_lst 
		self.label_lst = label_lst 
		self.smiles_lst = smiles_lst 
		self.icdcode_lst = icdcode_lst 
		self.criteria_lst = criteria_lst 

	def __len__(self):
		return len(self.nctid_lst)

	def __getitem__(self, index):
		return self.nctid_lst[index], self.label_lst[index], self.smiles_lst[index], self.icdcode_lst[index], self.criteria_lst[index]
	#### smiles_lst[index] is list of smiles

class Trial_Dataset_Complete(Trial_Dataset):
	def __init__(self, nctid_lst, status_lst, why_stop_lst, label_lst, phase_lst, 
					   diseases_lst, icdcode_lst, drugs_lst, smiles_lst, criteria_lst):
		Trial_Dataset.__init__(self, nctid_lst, label_lst, smiles_lst, icdcode_lst, criteria_lst)
		self.status_lst = status_lst 
		self.why_stop_lst = why_stop_lst 
		self.phase_lst = phase_lst 
		self.diseases_lst = diseases_lst 
		self.drugs_lst = drugs_lst 

	def __getitem__(self, index):
		return self.nctid_lst[index], self.status_lst[index], self.why_stop_lst[index], self.label_lst[index], self.phase_lst[index], \
			   self.diseases_lst[index], self.icdcode_lst[index], self.drugs_lst[index], self.smiles_lst[index], self.criteria_lst[index]



class ADMET_Dataset(data.Dataset):
	def __init__(self, smiles_lst, label_lst):
		self.smiles_lst = smiles_lst 
		self.label_lst = label_lst 
	
	def __len__(self):
		return len(self.smiles_lst)

	def __getitem__(self, index):
		return self.smiles_lst[index], self.label_lst[index]

def admet_collate_fn(x):
	smiles_lst = [i[0] for i in x]
	label_vec = default_collate([int(i[1]) for i in x])  ### shape n, 
	return [smiles_lst, label_vec]

def smiles_txt_to_lst(text):
	"""
		"['CN[C@H]1CC[C@@H](C2=CC(Cl)=C(Cl)C=C2)C2=CC=CC=C12', 'CNCCC=C1C2=CC=CC=C2CCC2=CC=CC=C12']" 
	"""
	text = text[1:-1]
	lst = [i.strip()[1:-1] for i in text.split(',')]
	return lst 

def icdcode_text_2_lst_of_lst(text):
	text = text[2:-2]
	lst_lst = []
	for i in text.split('", "'):
		i = i[1:-1]
		lst_lst.append([j.strip()[1:-1] for j in i.split(',')])
	return lst_lst 

def trial_collate_fn(x):
	nctid_lst = [i[0] for i in x]     ### ['NCT00604461', ..., 'NCT00788957'] 
	label_vec = default_collate([int(i[1]) for i in x])  ### shape n, 
	smiles_lst = [smiles_txt_to_lst(i[2]) for i in x]
	icdcode_lst = [icdcode_text_2_lst_of_lst(i[3]) for i in x]
	criteria_lst = [protocol2feature(i[4], sentence2vec) for i in x]
	return [nctid_lst, label_vec, smiles_lst, icdcode_lst, criteria_lst]

def trial_complete_collate_fn(x):
	nctid_lst = [i[0] for i in x]     ### ['NCT00604461', ..., 'NCT00788957'] 
	status_lst = [i[1] for i in x]
	why_stop_lst = [i[2] for i in x]
	label_vec = default_collate([int(i[3]) for i in x])  ### shape n, 
	phase_lst = [i[4] for i in x]
	diseases_lst = [i[5] for i in x]
	icdcode_lst = [icdcode_text_2_lst_of_lst(i[6]) for i in x]
	drugs_lst = [i[7] for i in x]
	smiles_lst = [smiles_txt_to_lst(i[8]) for i in x]
	criteria_lst = [protocol2feature(i[9], sentence2vec) for i in x]
	return [nctid_lst, status_lst, why_stop_lst, label_vec, phase_lst, diseases_lst, icdcode_lst, drugs_lst, smiles_lst, criteria_lst]

def csv_three_feature_2_dataloader(csvfile, shuffle, batch_size):
	with open(csvfile, 'r') as csvfile:
		rows = list(csv.reader(csvfile, delimiter=','))[1:]
	## nctid,status,why_stop,label,phase,diseases,icdcodes,drugs,smiless,criteria
	nctid_lst = [row[0] for row in rows]
	label_lst = [row[3] for row in rows]
	icdcode_lst = [row[6] for row in rows]
	drugs_lst = [row[7] for row in rows]
	smiles_lst = [row[8] for row in rows]
	criteria_lst = [row[9] for row in rows] 
	dataset = Trial_Dataset(nctid_lst, label_lst, smiles_lst, icdcode_lst, criteria_lst)
	data_loader = data.DataLoader(dataset, batch_size = batch_size, shuffle = shuffle, collate_fn = trial_collate_fn)
	return data_loader

def csv_three_feature_2_complete_dataloader(csvfile, shuffle, batch_size):
	with open(csvfile, 'r') as csvfile:
		rows = list(csv.reader(csvfile, delimiter=','))[1:]	
	nctid_lst = [row[0] for row in rows]
	status_lst = [row[1] for row in rows]
	why_stop_lst = [row[2] for row in rows]
	label_lst = [row[3] for row in rows]
	phase_lst = [row[4] for row in rows]
	diseases_lst = [row[5] for row in rows]
	icdcode_lst = [row[6] for row in rows]
	drugs_lst = [row[7] for row in rows]
	smiles_lst = [row[8] for row in rows]
	new_drugs_lst, new_smiles_lst = [], []
	criteria_lst = [row[9] for row in rows] 
	dataset = Trial_Dataset_Complete(nctid_lst, status_lst, why_stop_lst, label_lst, phase_lst, 
					   				 diseases_lst, icdcode_lst, drugs_lst, smiles_lst, criteria_lst)
	data_loader = data.DataLoader(dataset, batch_size = batch_size, shuffle = shuffle, collate_fn = trial_complete_collate_fn)
	return data_loader 






def smiles_txt_to_2lst(smiles_txt_file):
	with open(smiles_txt_file, 'r') as fin:
		lines = fin.readlines() 
	smiles_lst = [line.split()[0] for line in lines]
	label_lst = [int(line.split()[1]) for line in lines]
	return smiles_lst, label_lst 

def generate_admet_dataloader_lst(batch_size):
	datafolder = "data/ADMET/cooked/"
	name_lst = ["absorption", 'distribution', 'metabolism', 'excretion', 'toxicity']
	dataloader_lst = []
	for i,name in enumerate(name_lst):
		train_file = os.path.join(datafolder, name + '_train.txt')
		test_file = os.path.join(datafolder, name +'_valid.txt')
		train_smiles_lst, train_label_lst = smiles_txt_to_2lst(train_file)
		test_smiles_lst, test_label_lst = smiles_txt_to_2lst(test_file)
		train_dataset = ADMET_Dataset(smiles_lst = train_smiles_lst, label_lst = train_label_lst)
		test_dataset = ADMET_Dataset(smiles_lst = test_smiles_lst, label_lst = test_label_lst)
		train_dataloader = data.DataLoader(train_dataset, batch_size = batch_size, shuffle=True)
		test_dataloader = data.DataLoader(test_dataset, batch_size = batch_size, shuffle = False)
		dataloader_lst.append((train_dataloader, test_dataloader))
	return dataloader_lst 

# ## x is a list, len(x)=batch_size, x[i] is tuple, len(x[0])=5  
# def mpnn_feature_collate_func(x): 
# 	return [torch.cat([x[j][i] for j in range(len(x))], 0) for i in range(len(x[0]))]


# def mpnn_collate_func(x):
# 	#print("len(x) is ", len(x)) ## batch_size 
# 	#print("len(x[0]) is ", len(x[0])) ## 3--- data_process_loader.__getitem__ 
# 	mpnn_feature = [i[0] for i in x]
# 	#print("len(mpnn_feature)", len(mpnn_feature), "len(mpnn_feature[0])", len(mpnn_feature[0]))
# 	mpnn_feature = mpnn_feature_collate_func(mpnn_feature)
# 	from torch.utils.data.dataloader import default_collate
# 	x_remain = [i[1:] for i in x]
# 	x_remain_collated = default_collate(x_remain)
# 	return [mpnn_feature] + x_remain_collated

























import warnings
import pdb
import json
import pickle
import dill
from collections import defaultdict
import torch

from torch.utils.data import Dataset, DataLoader
import numpy as np

class Voc(object):
    def __init__(self):
        self.idx2word = {}
        self.word2idx = {}
    
    def __len__(self):
        return len(self.idx2word.keys())

    def add_sentence(self, sentence):
        for word in sentence:
            if word not in self.word2idx:
                self.idx2word[len(self.word2idx)] = word
                self.word2idx[word] = len(self.word2idx)

class SequencePatientBase(Dataset):
    '''
    Load sequential patient inputs for longitudinal patient records generation.
    Parameters
    ----------
    data: dict
        A dict contains patient data in sequence and/or in tabular.
        Given dict:
            {
                'x': np.ndarray or pd.DataFrame
                    Static patient features in tabular form, typically those baseline information.
                'v': list or np.ndarray
                    Patient visit sequence in dense format or in tensor format (depends on the model input requirement.)
                    If in dense format, it is like [[c1,c2,c3],[c4,c5],...], with shape [n_patient, NA, NA];
                    If in tensor format, it is like [[0,1,1],[1,1,0],...] (multi-hot encoded), 
                        with shape [n_patient, max_num_visit, max_num_event].
                
                'y': np.ndarray or pd.Series
                    Target label for each patient if making risk detection, with shape [n_patient, n_class];
                    Target label for each visit if making some visit-level prediction with shape [n_patient, NA, n_class].
            }
    
    metadata: dict (optional)
        A dict contains configuration of input patient data.
        metadata:
            {
                'voc': dict[Voc]
                    Vocabulary contains the event index to the exact event name, has three keys in general:
                    'diag', 'med', 'prod', corresponding to diagnosis, medication, and procedure.
                    ``Voc`` object should have two functions: `idx2word` and `word2idx`.
                
                'visit': dict[str]
                    a dict contains the format of input data for processing input visit sequences.
                    `visit`: {
                        'mode': 'tensor' or 'dense',
                        'order': list[str] (required when `mode='tensor'`)
                    },
                'label': dict[str]
                    a dict contains the format of input data for processing input labels.
                    `label`: {
                        'mode': 'tensor' or 'dense',
                    }
                
                'max_visit': int
                    the maximum number of visits considered when building tensor inputs, ignored
                    when visit mode is dense.
            }
    '''
    visit = None
    feature = None
    label = None
    max_visit = None
    visit_voc_size = None
    visit_order = None

    metadata = {
        'voc': {},
        
        'visit':{
            'mode': 'dense',
            'order': ['diag', 'prod', 'med'],
            },

        'label':{
            'mode': 'tensor',
            },

        'max_visit': 20,
    }

    def __init__(self, data, metadata=None) -> None:
        # parse metadata
        self._parse_metadata(metadata=metadata)

        # get input data
        self._parse_inputdata(inputs=data)

    def __getitem__(self, index):
        return_data = {}
        if self.visit is not None:
            visits = self.visit[index]
            if self.metadata['visit']['mode'] == 'tensor':
                visit_ts = self._dense_visit_to_tensor(visits) # return a dict with keys corresponding to order
                return_data['v'] = visit_ts
            else:
                visit_dict = self._parse_dense_visit_with_order(visits) # return a dict with keys corresponding to otder
                return_data['v'] = visit_dict
        
        if self.feature is not None:
            return_data['x'] = self.feature[index]
            
        if self.label is not None:
            return_data['y'] = self.label[index]
        
        return return_data

    def __len__(self):
        return len(self.visit)

    def _get_voc_size(self):
        order = self.metadata['visit']['order']
        vocs = self.metadata['voc']
        voc_size = []
        for order_ in order:
            voc_size.append(
                len(vocs[order_])
            )
        self.visit_voc_size = voc_size

    def _read_pickle(self, file_loc):
        return dill.load(open(file_loc, 'rb'))
    
    def _parse_metadata(self, metadata):
        if metadata is not None: 
            for k,v in metadata.items():
                if isinstance(v, dict):
                    self.metadata[k].update(v)
                else:
                    self.metadata[k] = v
        metadata = self.metadata

        if 'voc' in metadata:
            voc = metadata['voc']

            if 'diag' in voc: self.diag_voc = voc['diag']
            if 'prod' in voc: self.prod_voc = voc['prod']
            if 'med' in voc: self.med_voc = voc['med']
        
        if metadata['visit']['mode'] == 'tensor':
            self._get_voc_size()
        
        if 'order' in metadata['visit']:
            self.visit_order = metadata['visit']['order']

        if 'max_visit' in metadata:
            self.max_visit = metadata['max_visit']

    def _parse_inputdata(self, inputs):
        if 'x' in inputs: self.feature = inputs['x']
        if 'v' in inputs: self.visit = inputs['v']
        if 'y' in inputs: self.label = inputs['y']

    def _dense_visit_to_tensor(self, visits):
        res = {}
        for i,o in enumerate(self.visit_order):
            res[o] = np.zeros((self.max_visit, self.visit_voc_size[i]), dtype=int)

        for i, visit in enumerate(visits):
            # clip if the max visit is larger than self.max_visit
            if i >= self.max_visit: break
            
            for j, o in enumerate(self.visit_order):
                res[o][i, visit[j]] = 1
        return res
    
    def _parse_dense_visit_with_order(self, visits):
        return_data = defaultdict(list)
        order_list = self.metadata['visit']['order']
        for visit in visits:
            for i, o in enumerate(order_list):
                return_data[o].append(visit[i])
        return return_data


class SequencePatient(SequencePatientBase):
    '''
    Load sequential patient inputs for longitudinal patient records generation.
    Parameters
    ----------
    data: dict
        A dict contains patient data in sequence and/or in tabular.
        Given dict:
            {
                'x': np.ndarray or pd.DataFrame
                    Static patient features in tabular form, typically those baseline information.
                'v': list or np.ndarray
                    Patient visit sequence in dense format or in tensor format (depends on the model input requirement.)
                    If in dense format, it is like [[c1,c2,c3],[c4,c5],...], with shape [n_patient, NA, NA];
                    If in tensor format, it is like [[0,1,1],[1,1,0],...] (multi-hot encoded), 
                        with shape [n_patient, max_num_visit, max_num_event].
                
                'y': np.ndarray or pd.Series
                    Target label for each patient if making risk detection, with shape [n_patient, n_class];
                    Target label for each visit if making some visit-level prediction with shape [n_patient, NA, n_class].
            }
    
    metadata: dict (optional)
        A dict contains configuration of input patient data.
        metadata:
            {
                'voc': dict[Voc]
                    Vocabulary contains the event index to the exact event name, has three keys in general:
                    'diag', 'med', 'prod', corresponding to diagnosis, medication, and procedure.
                    ``Voc`` object should have two functions: `idx2word` and `word2idx`.
                
                'visit': dict[str]
                    a dict contains the format of input data for processing input visit sequences.
                    `visit`: {
                        'mode': 'tensor' or 'dense',
                        'order': list[str] (required when `mode='tensor'`)
                    },
                'label': dict[str]
                    a dict contains the format of input data for processing input labels.
                    `label`: {
                        'mode': 'tensor' or 'dense',
                    }
                
                'max_visit': int
                    the maximum number of visits considered when building tensor inputs, ignored
                    when visit mode is dense.
            }
    '''
    visit = None
    feature = None
    label = None
    max_visit = None
    visit_voc_size = None
    visit_order = None

    metadata = {
        'voc': {},
        
        'visit':{
            'mode': 'dense',
            'order': ['diag', 'prod', 'med'],
            },

        'label':{
            'mode': 'tensor',
            },

        'max_visit': 20,
    }

    def __init__(self, data, metadata=None) -> None:
        super().__init__(data=data, metadata=metadata)


import pdb
from collections import defaultdict

from pytrial.data.patient_data import SequencePatientBase
from pytrial.data.trial_data import TrialDatasetBase
from pytrial.data.patient_data import SeqPatientCollator

class PatientData(SequencePatientBase):
    '''
    Load sequence patient EHR data for patient-trial matching.
    '''
    def __init__(self, data, metadata=None) -> None:
        super().__init__(data, metadata)

class TrialData(TrialDatasetBase):
    '''
    Load trial dataset with eligibility criteria embedding.
    Each corresponds to a set matched patient indices for training
    patient-trial matching models.

    Parameters
    ----------
    data: pd.DataFrame
        A dataframe contains trial information including a column named `criteria` and 
        a column named `label` which contains a set of patient indices as matched.
    '''
    def __init__(self, data, encode_ec=False):
        super().__init__(data)
        # get criteria-level embeddings
        # stored in self.criteria_embedding as a matrix
        # self.df['inclusion_criteria_index'] and self.df['exclusion_criteria_index'] have the corresponding 
        # criteria indices for each trial.
        if encode_ec:
            self.get_ec_sentence_embedding()

    def __getitem__(self, index):
        # only get EC index and embeddings
        row = self.df.iloc[index]
        inc_ec_index = row['inclusion_criteria_index']
        if len(inc_ec_index) == 0: inc_ec_index.append(0)
        exc_ec_index = row['exclusion_criteria_index']
        if len(exc_ec_index) == 0: exc_ec_index.append(0) # pad
        output = {'inc_ec_index':inc_ec_index, 'exc_ec_index':exc_ec_index}
        
        inc_ec_emb = self.inc_ec_embedding[inc_ec_index]
        exc_ec_emb = self.exc_ec_embedding[exc_ec_index]
        nct_id = row['nctid']

        output.update(
            {
            'inc_ec_emb':inc_ec_emb,
            'exc_ec_emb':exc_ec_emb,
            'nct_id':nct_id,
            }
        )

        return output

class TrialCollator:
    '''
    Support the collation of trial records for patient-trial matching.
    '''
    def __call__(self, inputs):
        output = defaultdict(list)
        for x in inputs:
            output['inc_ec_index'].append(x['inc_ec_index'])
            output['exc_ec_index'].append(x['exc_ec_index'])
            output['inc_ec_emb'].append(x['inc_ec_emb'])
            output['exc_ec_emb'].append(x['exc_ec_emb'])
            output['nct_id'].append(x['nct_id'])
        return output

class PatientCollator(SeqPatientCollator):
    '''
    Support the collation of sequential patient EHR records for patient-trial matching.
    '''
    def __call__(self, inputs):
        '''
        Paramters
        ---------
        inputs = {
            'v': {
                'diag': list[np.ndarray],
                'prod': list[np.ndarray],
                'med': list[np.ndarray],
                },

            'x': list[np.ndarray],

            'y': list[np.ndarray]
        }

        Returns
        -------
        outputs = {
            'v':{ # visit event sequence
                'diag': tensor or list[tensor],
                'prod': tensor or list[tensor],
                'med': tensor or list[tensor],
            },
            'x': tensor, # static features
            'y': tensor or list, # patient-level label in tensor or visit-level label in list[tensor]
        }
        '''
        # init output dict
        return_data = defaultdict(list)
        return_data['v'] = defaultdict(list)

        for input in inputs:
            for k, v in input.items():
                if k == 'v': # visit seq
                    for key, value in v.items():
                        return_data['v'][key].append(value)
                else: # feature and label
                    return_data[k].append(v)

        # processing all
        if self.is_tensor_visit:
            self._parse_tensor_visit(return_data)
        
        if self.is_tensor_label:
            self._parse_tensor_label(return_data)

        if 'x' in input:
            self._parse_tensor_feature(return_data)

        return return_data


import torch
import torch.nn as nn
from torch.cuda.amp import autocast
from torch.utils.data import DataLoader
from pytrial.utils.trainer import Trainer

class SiteSelectTrainer(Trainer):
    '''
    Subclass the original trainer and provide specific evaluation functions and test_dataloader for sequence simulation.
    '''
    def prepare_input(self, data):
        return self.model._prepare_input(data)
    
    def train_one_iteration(self, 
        max_grad_norm=None,
        warmup_steps=None,
        use_amp=None, 
        scaler=None,
        train_loss_dict=None):
        skip_scheduler = False
        num_train_objectives = len(self.train_dataloader)
        for train_idx in range(num_train_objectives):
            data_iterator = self.data_iterators[train_idx]
            loss_model = self.loss_models[train_idx]
            loss_model.zero_grad()
            loss_model.train()
            optimizer = self.optimizers[train_idx]
            scheduler = self.schedulers[train_idx]
            try:
                data = next(data_iterator)
            except StopIteration:
                data_iterator = iter(self.train_dataloader[train_idx])
                self.data_iterators[train_idx] = data_iterator
                data = next(data_iterator)
                
            inputs = self.prepare_input(data)
            if use_amp:
                with autocast():
                    loss_return = loss_model(inputs)
                    loss = loss_return['loss_value']
            else:
                loss_return = loss_model(inputs)
                loss = loss_return['loss_value']

            # update for a single sample's all visits
            if use_amp:
                scale_before_step = scaler.get_scale()
                scaler.scale(loss).backward()
                scaler.unscale_(optimizer)
                torch.nn.utils.clip_grad_norm_(loss_model.parameters(), max_grad_norm)
                scaler.step(optimizer)
                scaler.update()
                skip_scheduler = scaler.get_scale() != scale_before_step
            else:
                loss.backward()
                torch.nn.utils.clip_grad_norm_(loss_model.parameters(), max_grad_norm)
                optimizer.step()

            train_loss_dict[train_idx].append(loss.item())
        
        if use_amp:
            skip_scheduler = scaler.get_scale() != scale_before_step

        if not skip_scheduler and warmup_steps > 0:
            scheduler.step()

    def _build_optimizer(self,
        loss_models,
        warmup_steps,
        num_train_steps,
        optimizer_param,
        weight_decay,
        scheduler=None,
        ):
        optimizer_class = torch.optim.AdamW
        optimizers = []
        schedulers = []
        for loss_model in loss_models:
            param_optimizer = list(loss_model.named_parameters())
            no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
            optimizer_grouped_parameters = [
                {'params': [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)], 'weight_decay': weight_decay},
                {'params': [p for n, p in param_optimizer if any(nd in n for nd in no_decay)], 'weight_decay': 0.0}
            ]
            optimizer = optimizer_class(optimizer_grouped_parameters, **optimizer_param)
            if warmup_steps > 0:
                scheduler_obj = self._build_scheduler(optimizer, scheduler, warmup_steps=warmup_steps, t_total=num_train_steps)
            else:
                scheduler_obj = None
            optimizers.append(optimizer)
            schedulers.append(scheduler_obj)
        return optimizers, schedulers


import abc
from collections import defaultdict
import pdb
import os
import json
import math

import torch
from torch import nn
import numpy as np

from pytrial.utils.check import check_model_dir
from pytrial.utils.check import make_dir_if_not_exist
from ..data import TrialData, PatientData

class PatientTrialMatchBase(abc.ABC):
    '''Abstract class for all sequential patient data simulations.
    
    Parameters
    ----------
    experiment_id: str, optional (default = 'test')
        The name of current experiment.
    '''
    training=False
    @abc.abstractmethod
    def __init__(self, experiment_id='test'):
        check_model_dir(experiment_id)
        self.checkout_dir = os.path.join('./experiments_records', experiment_id,
                                         'checkpoints')
        self.result_dir = os.path.join('./experiments_records', experiment_id,
                                       'results')
        make_dir_if_not_exist(self.checkout_dir)
        make_dir_if_not_exist(self.result_dir)
    
    @abc.abstractmethod
    def fit(self, train_data):
        '''
        Fit function needs to be implemented after subclass.

        Parameters
        ----------
        train_data: Any
            Training data.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def predict(self, test_data):
        '''
        Prediction function needs to be implemented after subclass.

        Parameters
        ----------
        test_data: Any
            Testing data.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def save_model(self, output_dir):
        '''
        Save the model to disk, needs to be implemented after subclass.

        Parameters
        ----------
        output_dir: str
            The path to the output directory.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def load_model(self, checkpoint):
        '''
        Load the pretrained model from disk, needs to be implemented after subclass.
        
        Parameters
        ----------
        checkpoint: str
            The path to the checkpoint file.
        '''
        raise NotImplementedError
    
    def train(self, mode=True):
        '''
        Swith the model to the `training` mode. Work samely as `model.train()` in pytorch.

        Parameters
        ----------
        mode: bool, optional (default = True)
            If True, switch to the `training` mode.
        '''
        self.training = mode
        self.model.train()
        return self
    
    def eval(self, mode=False):
        '''
        Swith the model to the `validation` mode. Work samely as `model.eval()` in pytorch.

        Parameters
        ----------
        mode: bool, optional (default = False)
            If False, switch to the `validation` mode.
        '''
        self.training = mode
        self.model.eval()
        return self

    def _input_data_check(self, inputs):
        assert 'patient' in inputs, 'Do not find patient data in inputs!'
        assert 'trial' in inputs, 'Do not find trial data in inputs!'
        assert isinstance(inputs['trial'], TrialData), 'Input trial data is not a `trial_patient_match.data.TrialData` instance!'
        assert isinstance(inputs['patient'], PatientData), 'Input patient data is not a `trial_patient_match.data.PatientData` instance!'
    
    def _save_checkpoint(self, state,
                        epoch_id=0,
                        is_best=False,
                        output_dir=None,
                        filename='checkpoint.pth.tar'):
        if output_dir is None:
            output_dir = self.checkout_dir

        if epoch_id < 1:
            filepath = os.path.join(output_dir, 'latest.' + filename)
        elif is_best:
            filepath = os.path.join(output_dir, 'best.' + filename)
        else:
            filepath = os.path.join(self.checkout_dir,
                                    str(epoch_id) + '.' + filename)
        torch.save(state, filepath)

    def _save_config(self, config, output_dir=None):
        if output_dir is None:
            output_dir = self.checkout_dir

        temp_path = os.path.join(output_dir, 'config.json')
        if os.path.exists(temp_path):
            os.remove(temp_path)
        with open(temp_path, 'w', encoding='utf-8') as f:
            f.write(
                json.dumps(config, indent=4)
            )

    def _load_config(self, checkpoint=None):
        '''
        Load model config from the given directory.
        Parameters
        ----------
        checkpoint: str
            The given filepath (e.g., ./checkpoint/config.json)
            to load the model config.
        '''
        if checkpoint is None:
            temp_path = os.path.join(checkpoint, 'config.json')
        else:
            temp_path = checkpoint

        assert os.path.exists(temp_path), 'Cannot find `config.json` under {}'.format(self.checkout_dir)
        with open(temp_path, 'r') as f:
            config = json.load(f)
        return config

    def _tuple_result_to_dict(self, pred_res):
        res = defaultdict(list)
        for nctid, pred in pred_res:
            if isinstance(nctid, list):
                for i,nctid_ in enumerate(nctid):
                    res[nctid_] = pred[i].tolist()
            else:
                res[nctid] = pred.flatten().tolist()
        return res

    def _translate_dense_visits_to_sparse(self, visits):
        total_vocab_size = sum(self.config['vocab_size'])
        num_visits = len(visits[self.config['order'][0]])
        outputs = np.zeros((num_visits, total_vocab_size))

        for i, o in enumerate(self.config['order']):
            for j in range(num_visits):
                raw = visits[o][j]
                if isinstance(raw, torch.Tensor): raw = raw.detach().cpu().numpy()
                if i > 0: 
                    voc_size = sum(self.config['vocab_size'][:i-1])
                    if isinstance(raw, list):
                        raw = [r + voc_size for r in raw]
                    else:
                        raw += voc_size
                outputs[j, raw] = 1

        return outputs

    def _match_trial_for_patients(self, pred_inc, criteria_type='inc'):
        pred_label_ec = torch.max(pred_inc, -1)[1] # 1，num_patients, num_ec
        num_ec = pred_label_ec.shape[-1]

        # prediction logits
        # 0 is unmatch
        # 1 is match
        # 2 is unknown

        if criteria_type == 'inc':
            # prediction
            pred_label_trial_inc = (pred_label_ec == 1).sum(-1) # 1 is match
            pred_label_trial_inc[pred_label_trial_inc<num_ec] = 0 # does not satisfy all criteria, not matched
            pred_label_trial_inc[pred_label_trial_inc==num_ec] = 1 # 1, num_patients
            return pred_label_trial_inc

        else:
            pred_label_trial_exc = (pred_label_ec == 0).sum(-1) # 0 is unmatch, do not match all exclusion criteria then sum == num_ec
            pred_label_trial_exc[pred_label_trial_exc<num_ec] = 0 # has more than 1 exclusion criteria satisfied, not matched
            pred_label_trial_exc[pred_label_trial_exc==num_ec] = 1 # 1, num_patients
            return pred_label_trial_exc


from cmath import isnan
import pdb
import warnings

from transformers.models.bart.configuration_bart import BartConfig
from transformers.models.bart.modeling_bart import BartModel, BartPretrainedModel
from transformers.models.bart.modeling_bart import shift_tokens_right
import torch
from torch import nn

from .generator import EHRGenerationMixin
from .modeling_config import EHRBartOutput
from .modeling_config import DataTokenizer, ModelTokenizer
from .modeling_config import EHRBartConfig
from .modeling_promptbart import PromptBartModel
from . import constants

class BartForEHRSimulation(BartPretrainedModel, EHRGenerationMixin):
    def __init__(self, config: EHRBartConfig, model_tokenizer: ModelTokenizer=None, data_tokenizer: DataTokenizer=None):
        super().__init__(config)

        config.is_decoder = False
        config.is_encoder_decoder = True # use both the inputs for encoders and decoders

        self.model = PromptBartModel(config)

        self.register_buffer("final_logits_bias", torch.zeros((1, self.model.shared.num_embeddings)))

        self.lm_head = {}
        for key in model_tokenizer.tokenizer_dict.keys():
            self.lm_head[key] =  nn.Linear(
                config.d_model, config.__dict__[key], bias=False,
            )
        self.lm_head = nn.ModuleDict(self.lm_head)

        self.init_weights()

        # set embedding vocab new size
        self.resize_token_embeddings(config.data_tokenizer_num_vocab)

        # for specific modal generation
        self.model_tokenizer = model_tokenizer
        self.data_tokenizer = data_tokenizer

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        decoder_input_ids=None,
        decoder_attention_mask=None,
        x_num = None,
        x_cat = None,
        head_mask=None,
        decoder_head_mask=None,
        cross_attn_head_mask=None,
        encoder_outputs=None,
        past_key_values=None,
        inputs_embeds=None,
        decoder_inputs_embeds=None,
        labels=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=True,
        label_mask=None,
        code_type=None,
    ):
        r"""This forward is specified for *ALL* possible words in the vocabulary. It's same as the BartForConditionalGeneration.
        labels (:obj:`torch.LongTensor` of shape :obj:`(batch_size, sequence_length)`, `optional`):
            Labels for computing the masked language modeling loss. Indices should either be in ``[0, ...,
            config.vocab_size]`` or -100 (see ``input_ids`` docstring). Tokens with indices set to ``-100`` are ignored
            (masked), the loss is only computed for the tokens with labels in ``[0, ..., config.vocab_size]``.

        Returns:
        """
        code_types = list(self.model_tokenizer.tokenizer_dict.keys())
        assert code_type in code_types

        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if labels is not None:
            if decoder_input_ids is None:
                decoder_input_ids = shift_tokens_right(
                    labels, self.config.pad_token_id, self.config.decoder_start_token_id,
                )
        
        # get the outputs from both encoder and decoder of BartModel
        # if only input_ids are given and no decoder_input_ids is given
        # the model will shift input_ids to get the decoder input ids during forward
        outputs = self.model(
            input_ids,
            attention_mask=attention_mask, # 0 for"masked", 1 for "Not masked"
            decoder_input_ids=decoder_input_ids,
            x_num=x_num,
            x_cat=x_cat,
            encoder_outputs=encoder_outputs,
            decoder_attention_mask=decoder_attention_mask,
            head_mask=head_mask,
            decoder_head_mask=decoder_head_mask,
            cross_attn_head_mask=cross_attn_head_mask,
            past_key_values=past_key_values,
            inputs_embeds=inputs_embeds,
            decoder_inputs_embeds=decoder_inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        # for diagnosis/procedure/lab tests/medications
        logits = self.lm_head[code_type](outputs[0])

        loss = None
        perplexity = None
        if labels is not None:
            encoded_labels = self.model_tokenizer.encode(labels, code_type)
            encoded_labels = encoded_labels - self.model_tokenizer.label_offset
            encoded_labels[encoded_labels < 0] = -100 # ignore special tokens when computing losses

            if (x_num is not None or x_cat is not None) \
                and (self.model.encoder_conditional_prompt is not None) \
                and (self.model.decoder_conditional_prompt is not None):
                # adjust label mask if context conditional prompt is used

                n_num = x_num.shape[1] if x_num is not None else 0
                n_cat = x_cat.shape[1] if x_cat is not None else 0
                num_label_offset = n_num + n_cat
                label_offset = torch.ones(encoded_labels.shape[0], num_label_offset) * -100
                label_offset = label_offset.long().to(labels.device)
                encoded_labels = torch.cat([label_offset, encoded_labels], dim=1)
                if label_mask is not None:
                    label_mask_offset = torch.zeros(encoded_labels.shape[0], num_label_offset)
                    label_mask_offset = label_mask_offset.to(label_mask.device)
                    label_mask = torch.cat([label_mask_offset, label_mask], dim=1)

            # compute loss
            loss_fct = nn.CrossEntropyLoss(ignore_index=-100)
            loss = loss_fct(logits.view(-1, self.lm_head[code_type].out_features), encoded_labels.view(-1))
            
            if label_mask is not None: # do evaluation, compute perplexity
                if encoded_labels[encoded_labels >= 0].shape[0] == 0:
                    perplexity = None
                else:
                    target = encoded_labels[label_mask.bool()]
                    mask_logits = logits[label_mask.bool()]

                    # debug: move to CPU see errors
                    # prob = torch.gather(mask_logits.softmax(1).cpu(), 1, target.unsqueeze(-1).cpu())

                    prob = torch.gather(mask_logits.softmax(1), 1, target.unsqueeze(-1))
                    nll = -torch.log(prob+constants.eps)
                    perplexity = nll.exp()
                    if torch.isnan(perplexity).any():
                        warnings.warn('Find NaN perplexity during the forward of PromptEHR model!')
                    perplexity = perplexity.median()
                
        if not return_dict:
            output = (logits,) + outputs[1:]
            return ((loss,) + output) if loss is not None else output

        # debug
        # if logits.isnan().sum() > 0:
        #     pdb.set_trace()

        return EHRBartOutput(
            loss=loss,
            logits=logits,
            past_key_values=outputs.past_key_values,
            decoder_hidden_states=outputs.decoder_hidden_states,
            decoder_attentions=outputs.decoder_attentions,
            cross_attentions=outputs.cross_attentions,
            encoder_last_hidden_state=outputs.encoder_last_hidden_state,
            encoder_hidden_states=outputs.encoder_hidden_states,
            encoder_attentions=outputs.encoder_attentions,
            perplexity=perplexity,
        )

    @staticmethod
    def _reorder_cache(past, beam_idx):
        reordered_past = ()
        for layer_past in past:
            # cached cross_attention states don't have to be reordered -> they are always the same
            reordered_past += (
                tuple(past_state.index_select(0, beam_idx) for past_state in layer_past[:2]) + layer_past[2:],
            )
        return reordered_past

    def prepare_inputs_for_generation(
        self,
        decoder_input_ids,
        past=None,
        attention_mask=None,
        head_mask=None,
        decoder_head_mask=None,
        cross_attn_head_mask=None,
        use_cache=None,
        encoder_outputs=None,
        **kwargs,
    ):
        # cut decoder_input_ids if past is used
        if past is not None:
            decoder_input_ids = decoder_input_ids[:, -1:]
        
        return_res = {
            "input_ids": None,  # encoder_outputs is defined. input_ids not needed
            "encoder_outputs": encoder_outputs,
            "past_key_values": past,
            "decoder_input_ids": decoder_input_ids,
            "attention_mask": attention_mask,
            "head_mask": head_mask,
            "decoder_head_mask": decoder_head_mask,
            "cross_attn_head_mask": cross_attn_head_mask,
            "use_cache": use_cache,  # change this to avoid caching (presumably for debugging)
        }

        if 'x_cat' in kwargs: return_res['x_cat'] = kwargs['x_cat']
        if 'x_num' in kwargs: return_res['x_num'] = kwargs['x_num']

        return return_res

    def prepare_decoder_input_ids_from_labels(self, labels: torch.Tensor):
        return shift_tokens_right(labels, self.config.pad_token_id, self.config.decoder_start_token_id)

    def get_encoder(self):
        return self.model.get_encoder()

    def get_decoder(self):
        return self.model.get_decoder()

    def resize_token_embeddings(self, new_num_tokens: int) -> nn.Embedding:
        new_embeddings = super().resize_token_embeddings(new_num_tokens)
        self._resize_final_logits_bias(new_num_tokens)
        return new_embeddings

    def _resize_final_logits_bias(self, new_num_tokens: int) -> None:
        old_num_tokens = self.final_logits_bias.shape[-1]
        if new_num_tokens <= old_num_tokens:
            new_bias = self.final_logits_bias[:, :new_num_tokens]
        else:
            extra_bias = torch.zeros((1, new_num_tokens - old_num_tokens), device=self.final_logits_bias.device)
            new_bias = torch.cat([self.final_logits_bias, extra_bias], dim=1)
        self.register_buffer("final_logits_bias", new_bias)

    def get_output_embeddings(self):
        return self.lm_head

    def set_output_embeddings(self, new_embeddings):
        self.lm_head = new_embeddings

    def get_prompt_encoder(self):
        return self.model.encoder_conditional_prompt
    
    def get_prompt_decoder(self):
        return self.model.decoder_conditional_prompt



import abc
import os
import json
import pandas as pd

from pytrial.utils.check import check_model_dir
from pytrial.utils.check import make_dir_if_not_exist
from pytrial.data.patient_data import TabularPatientBase


class TabularSimulationBase(abc.ABC):
    '''Abstract class for all tabular simulations
    based on tabular patient data.
    
    Parameters
    ----------
    experiment_id: str, optional (default = 'test')
        The name of current experiment.
    '''
    @abc.abstractmethod
    def __init__(self, experiment_id='trial_simulation.tabular'):
        check_model_dir(experiment_id)
        self.checkout_dir = os.path.join('./experiments_records', experiment_id,
                                         'checkpoints')
        self.result_dir = os.path.join('./experiments_records', experiment_id,
                                       'results')
        make_dir_if_not_exist(self.checkout_dir)
        make_dir_if_not_exist(self.result_dir)

    @abc.abstractmethod
    def fit(self, train_data):
        '''Fit the model to the given training data. Need to be implemented by subclass.

        Parameters
        ----------
        train_data: Any
            The training data.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def predict(self, number_of_predictions):
        '''Generate synthetic data after the model trained. Need to be implemented by subclass.

        Parameters
        ----------
        number_of_predictions: Any
            Number of synthetic data to be generated.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def save_model(self, output_dir):
        '''Save the model to the given directory. Need to be implemented by subclass.
        
        Parameters
        ----------
        output_dir: str
            The given directory to save the model.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def load_model(self, checkpoint):
        '''
        Load model from the given directory. Need to be implemented by subclass.
        
        Parameters
        ----------
        checkpoint: str
            The given filepath (e.g., ./checkpoint/model.pth)
        '''
        raise NotImplementedError

    def _input_data_check(self, inputs):
        assert (isinstance(inputs, TabularPatientBase) or isinstance(inputs, dict)), 'Wrong input type.'

    def _save_config(self, config, output_dir=None):
        '''
        Save model config to the given directory.

        Parameters
        ----------
        config: dict
            Hyperparameters and model infomation
            
        checkpoint: str
            The given filepath (e.g., ./checkpoint/config.json)
            to load the model config.
        '''
        if output_dir is None:
            output_dir = self.checkout_dir

        temp_path = os.path.join(output_dir, 'config.json')
        if os.path.exists(temp_path):
            os.remove(temp_path)
        with open(temp_path, 'w', encoding='utf-8') as f:
            f.write(
                json.dumps(config, indent=4)
            )

    def _load_config(self, checkpoint=None):
        '''
        Load model config from the given directory.

        Parameters
        ----------
        checkpoint: str
            The given filepath (e.g., ./checkpoint/config.json)
            to load the model config.
        '''
        if checkpoint is None:
            temp_path = os.path.join(checkpoint, 'config.json')
        else:
            temp_path = checkpoint

        assert os.path.exists(temp_path), 'Cannot find `config.json` under {}'.format(self.checkout_dir)
        with open(temp_path, 'r') as f:
            config = json.load(f)
        return config



'''
Provide an easy-to-acess function to load ready-to-use demo data
from './demo_data' folder.
'''
import pdb
import os
import dill
import wget

import pandas as pd

from .patient_data import TabularPatientBase
from ..utils.trial_utils import ClinicalTrials
from ..utils.tabular_utils import read_csv_to_df, load_table_config

TRIALSIM_DATA_URL = 'https://storage.googleapis.com/pytrial/TrialSim-data.xlsx'
SYNTHETIC_DATA_URL = 'https://github.com/RyanWangZf/PromptEHR/raw/main/demo_data/synthetic_ehr/data.pkl'
SEQ_TRIAL_PATIENT_URL = 'https://storage.googleapis.com/pytrial/seq_patient_nct00174655.zip'
TOP_URL = 'https://github.com/futianfan/clinical-trial-outcome-prediction/raw/main/data/'
TRIAL_PATIENT_URL = 'https://storage.googleapis.com/pytrial/demo_trial_patient_data.zip'
TRIAL_DOCUMENT_URL = 'https://storage.googleapis.com/pytrial/clinical_trials.csv'

__all__ = [
    'load_mimic_ehr_sequence',
    'load_synthetic_ehr_sequence',
    'load_trial_patient_sequence',
    'load_trial_patient_tabular',
    'load_trial_outcome_data',
    'load_trial_document_data',
]


def load_synthetic_ehr_sequence(input_dir=None, n_sample=None):
    '''
    Load synthetic EHR patient sequence data, which was generated by PromptEHR (https://arxiv.org/pdf/2211.01761.pdf).

    Parameters
    ----------
    input_dir: str
        The folder that stores the demo data. If None, we will download the demo data and save it
        to './demo_data/synthetic_ehr'. Make sure to remove this folder if it is empty.
    
    n_sample: int
        The number of samples we want to load. If None, all data will be loaded.
    '''
    if input_dir is None:
        input_dir = './demo_data/synthetic_ehr'

    if not os.path.exists(input_dir):
        os.makedirs(input_dir)
        url = SYNTHETIC_DATA_URL
        filename = wget.download(url, out=input_dir)
        print(f'Download synthetic EHRs to {input_dir}.')
    
    with open(os.path.join(input_dir,'data.pkl'), 'rb') as f:
        x = dill.load(f)
    
    # ZW: temporal solution to solve the typo
    x['cat_cardinalities'] = x.pop('cat_cardinalties')

    if n_sample is not None:
        # cut to get smaller demo data
        x['visit'] = x['visit'][:n_sample]
        x['y'] = x['y'][:n_sample]
        x['feature'] = x['feature'][:n_sample]

    return x


def load_mimic_ehr_sequence(input_dir=None, n_sample=None):
    '''
    Load EHR patient sequence data, which needs to be accessed via https://physionet.org/content/mimiciii/1.4/.

    Parameters
    ----------
    input_dir: str
        The folder that stores the demo data. If None, we will look for the demo data in
        './demo_data/demo_patient_sequence/ehr'.
    
    n_sample: int
        The number of samples we want to load. If None, all data will be loaded.
    '''
    if input_dir is None:
        input_dir = './demo_data/demo_patient_sequence/ehr'

    if not os.path.exists(input_dir):
        raise ValueError(f'Please download the MIMIC-III dataset and put it in {input_dir}.')

    visit = dill.load(open(os.path.join(input_dir, 'visits.pkl'), 'rb'))
    voc = dill.load(open(os.path.join(input_dir, 'voc.pkl'), 'rb'))

    # make some simple processing
    feature = pd.read_csv(os.path.join(input_dir, 'feature.csv'), index_col=0)
    label = feature['MORTALITY'].values
    x = feature[['AGE','GENDER','ETHNICITY']]
    tabx = TabularPatientBase(x)
    x = tabx.df.values # get processed patient features in matrix form

    if n_sample is not None:
        # cut to get smaller demo data
        visit = visit[:n_sample]
        label = label[:n_sample]
        x = x[:n_sample]

    n_num_feature = 1
    cat_cardinalities = []
    for i in range(n_num_feature, x.shape[1]):
        cat_cardinalities.append(len(list(set(x[:,i]))))

    return {
        'visit':visit,
        'voc':voc,
        'order':['diag','prod','med'],
        'mortality':label,
        'feature':x,
        'n_num_feature':n_num_feature,
        'cat_cardinalities':cat_cardinalities,
        }

def load_trial_patient_sequence(input_dir=None):
    '''
    Load synthetic sequential trial patient records.

    Parameters
    ----------
    input_dir: str
        The folder that stores the demo data. If None, we will download the demo data and save it
        to './demo_data/demo_patient_sequence/trial'. Make sure to remove this folder if it is empty.    
    '''
    if input_dir is None:
        input_dir = './demo_data/demo_patient_sequence/trial'
    
    if not os.path.exists(input_dir):
        os.makedirs(input_dir)
        url = SEQ_TRIAL_PATIENT_URL
        filename = wget.download(url, out=input_dir)
        # unzip filename
        import zipfile
        with zipfile.ZipFile(filename, 'r') as zip_ref:
            zip_ref.extractall(input_dir)
        print(f'\n Download trial patient sequence data to {input_dir}.')

    # load patient data
    print("#"*5+'Demo Data Folder'+"#"*5)
    print(os.listdir(input_dir))
    print("#"*20)
    visit = dill.load(open(os.path.join(input_dir,'visit.pkl'), 'rb'))
    vocs = dill.load(open(os.path.join(input_dir,'voc.pkl'), 'rb'))
    feature = pd.read_csv(os.path.join(input_dir, 'feature.csv'))
    v_stage = dill.load(open(os.path.join(input_dir,'visit_stage.pkl'), 'rb'))
    orders = list(vocs.keys())
    # data preprocessing
    label_relapse = feature['num relapse']
    label_mortality = feature['death'].values
    x = feature.drop(['num relapse','death','RUSUBJID'], axis=1)
    x['weight'] = x['weight'].replace({'>= 125':'125'}).astype(float)
    tabx = TabularPatientBase(x)
    x = tabx.df.values # get processed patient features in matrix form
    return {
        'feature':x,
        'visit':visit,
        'voc':vocs,
        'order':orders,
        'visit_stage':v_stage,
        'relapse':label_relapse,
        'mortality':label_mortality,
    }

def load_trial_outcome_data(input_dir=None, phase='I', split='train'):
    '''
    Load trial outcome prediction (TOP) benchmark data.

    Parameters
    ----------
    input_dir: str
        The folder that stores the demo data. If None, we will download the demo data and save it
        to './demo_data/demo_trial_data'. Make sure to remove this folder if it is empty.

    phase: {'I','II','III'}
        The phase of the trial data. Can be 'I', 'II', 'III'.
    
    split: {'train', 'test', 'valid'}
        The split of the trial data. Can be 'train', 'test', 'valid'.
    '''

    BENCHMARK_DATA_URL = 'https://storage.googleapis.com/pytrial/HINT-benchmark-data/hint_benchmark_dataset_w_date.zip'

    if input_dir is None:
        input_dir = './demo_data/demo_trial_outcome_data'

    if not os.path.exists(input_dir):
        os.makedirs(input_dir)
        # download the benchmark data
        wget.download(BENCHMARK_DATA_URL, out=input_dir)
        # unzip filename
        import zipfile
        with zipfile.ZipFile(os.path.join(input_dir, 'hint_benchmark_dataset_w_date.zip'), 'r') as zip_ref:
            zip_ref.extractall(input_dir)
        print(f'\n Download trial data to {input_dir}.')
    
    filename = 'phase_{}_{}.csv'.format(phase, split)
    filename = os.path.join(input_dir, filename)
    # load patient data
    df = pd.read_csv(filename)
    return {'data':df}

def load_trial_document_data(input_dir=None, 
    n_sample=None,
    source='preprocessed',
    date='20221001',
    ):
    '''
    Load trial document data obtained from ClinicalTrials.gov.

    Parameters
    ----------
    input_dir: str
        The folder that stores the demo data. If None, we will download the demo data and save it
        to ''./demo_data/demo_trial_document'. Make sure to remove this folder if it is empty.
    
    n_sample: int
        The number of samples we want to load. If None, all data will be loaded.
    
    source: {'clinicaltrials.gov', 'preprocessed'}
        The source of the data. If 'clinicaltrials.gov', we will download the raw data from
        that website and process it. If 'preprocessed', we will load the preprocessed data.

    date: str
        The date of the clinicaltrials.gov copy. Only valid when ``source='clinicaltrials.gov'``.
    '''
    if input_dir is None:
        input_dir = './demo_data/demo_trial_document'
    
    if not os.path.exists(input_dir):
        os.makedirs(input_dir)
    
    filepath = os.path.join(input_dir, 'clinical_trials.csv')
    label_filepath = os.path.join(input_dir, 'TrialSim-data.xlsx')
    
    if not os.path.exists(filepath):
        if source == 'clinicaltrials.gov':
            print("Downloading clinicaltrials.gov data...")
            # download demo data
            client = ClinicalTrials()
            client.download(date=date, output_dir=input_dir)
        else:
            # download the preprocessed data
            print("Downloading preprocessed clinical trial documents data (copy of 10/01/2022)...")
            url = TRIAL_DOCUMENT_URL
            import wget
            filepath = wget.download(url, out=os.path.join(input_dir, 'clinical_trials.csv'))

    if not os.path.exists(label_filepath):
        # download demo label
        import wget
        wget.download(TRIALSIM_DATA_URL, out=label_filepath)

    df = pd.read_csv(filepath, index_col=0)
    df_tr = df # all data
    df_val = pd.read_excel(label_filepath, index_col=0)

    df_v = pd.DataFrame({'nct_id':df_val.iloc[:,:11].to_numpy().flatten()})
    df_v = df_v.merge(df, on='nct_id', how='inner')
    df_tr = pd.concat([df_tr, df_v], axis=0).drop_duplicates()

    if n_sample is not None:
        # cut to get smaller demo data
        df_tr = df_tr.iloc[:n_sample]

    return {
        'x': df_tr,
        'fields':['title','intervention_name','disease','keyword'],
        'ctx_fields':['description','criteria'],
        'tag': 'nct_id',
        'x_val':df_val.iloc[:,:11],
        'y_val':df_val.iloc[:,11:],
    }

def load_trial_patient_tabular(input_dir=None):
    '''
    Load synthetic tabular trial patient records.

    Parameters
    ----------
    input_dir: str
        The folder that stores the demo data. If None, we will download the demo data and save it
        to './demo_data/demo_trial_patient_data'. Make sure to remove this folder if it is empty.    
    '''
    if input_dir is None:
        input_dir = './demo_data/demo_trial_patient_data'
    
    if not os.path.exists(input_dir):
        os.makedirs(input_dir)
        url = TRIAL_PATIENT_URL
        filename = wget.download(url, out=input_dir)
        # unzip filename
        import zipfile
        with zipfile.ZipFile(filename, 'r') as zip_ref:
            zip_ref.extractall(input_dir)
        print(f'\n Download trial patient data to {input_dir}.')
    
    # load patient data
    df = read_csv_to_df(os.path.join(input_dir, 'data_processed.csv'), index_col=0)
    table_config = load_table_config(input_dir)
    return {'data':df, 'metadata':table_config}



"""A set of utility functions to support parallel computation.

Code comes from: https://github.com/zzachw/PyHealth/blob/master/pyhealth/utils/utility_parallel.py
"""

from joblib import effective_n_jobs
import contextlib
import joblib
from tqdm import tqdm
from joblib import Parallel, delayed

import numpy as np
from torch import nn
from torch import Tensor, device


@contextlib.contextmanager
def tqdm_joblib(tqdm_object):
    """Context manager to patch joblib to report into tqdm progress bar given as argument"""

    class TqdmBatchCompletionCallback(joblib.parallel.BatchCompletionCallBack):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def __call__(self, *args, **kwargs):
            tqdm_object.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback
    try:
        yield tqdm_object
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_object.close()

# This elegent solution is from:
# https://stackoverflow.com/questions/24983493/tracking-progress-of-joblib-parallel-execution

def partition_estimators(n_estimators, n_jobs):
    """Private function used to partition estimators between jobs."""
    # Compute the number of jobs
    n_jobs = min(effective_n_jobs(n_jobs), n_estimators)

    # Partition estimators between jobs
    n_estimators_per_job = np.full(n_jobs, n_estimators // n_jobs,
                                   dtype=np.int)
    n_estimators_per_job[:n_estimators % n_jobs] += 1
    starts = np.cumsum(n_estimators_per_job)

    xdiff = [starts[n] - starts[n - 1] for n in range(1, len(starts))]

    # print("Split among workers default:", starts, xdiff)
    print('Split tasks into', n_jobs, 'cores...')
    return n_estimators_per_job.tolist(), [0] + starts.tolist(), n_jobs


def unfold_parallel(lists, n_jobs):
    """Internal function to unfold the results returned from the parallization
    Parameters
    ----------
    lists : list
        The results from the parallelization operations.
    n_jobs : optional (default=1)
        The number of jobs to run in parallel for both `fit` and
        `predict`. If -1, then the number of jobs is set to the
        number of cores.
    Returns
    -------
    result_list : list
        The list of unfolded result.
    """
    full_list = []
    for i in range(n_jobs):
        full_list.extend(lists[i])
    return full_list

def unwrap_model(model: nn.Module) -> nn.Module:
    """
    Recursively unwraps a model from potential containers (as used in distributed training).
    Args:
        model (`torch.nn.Module`): The model to unwrap.
    """
    # since there could be multiple levels of wrapping, unwrap recursively
    if hasattr(model, "module"):
        return unwrap_model(model.module)
    else:
        return model

def batch_to_device(batch, target_device: device):
    """
    send a pytorch batch to a device (CPU/GPU)
    """
    for key in batch:
        if isinstance(batch[key], Tensor):
            batch[key] = batch[key].to(target_device)
    return batch

'''
The following codes come from https://github.com/zhanghang1989/PyTorch-Encoding/blob/master/encoding/parallel.py.
'''
##+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
## Created by: Hang Zhang
## ECE Department, Rutgers University
## Email: zhang.hang@rutgers.edu
## Copyright (c) 2017
##
## This source code is licensed under the MIT-style license found in the
## LICENSE file in the root directory of this source tree
##+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import threading
import functools
import torch
from torch.autograd import Variable, Function
import torch.cuda.comm as comm
from torch.nn.parallel.data_parallel import DataParallel
from torch.nn.parallel.parallel_apply import get_a_var
from torch.nn.parallel._functions import ReduceAddCoalesced, Broadcast

torch_ver = torch.__version__[:3]

def allreduce(*inputs):
    """Cross GPU all reduce autograd operation for calculate mean and
    variance in SyncBN.
    """
    return AllReduce.apply(*inputs)

class AllReduce(Function):
    @staticmethod
    def forward(ctx, num_inputs, *inputs):
        ctx.num_inputs = num_inputs
        ctx.target_gpus = [inputs[i].get_device() for i in range(0, len(inputs), num_inputs)]
        inputs = [inputs[i:i + num_inputs]
                 for i in range(0, len(inputs), num_inputs)]
        # sort before reduce sum
        inputs = sorted(inputs, key=lambda i: i[0].get_device())
        results = comm.reduce_add_coalesced(inputs, ctx.target_gpus[0])
        outputs = comm.broadcast_coalesced(results, ctx.target_gpus)
        return tuple([t for tensors in outputs for t in tensors])

    @staticmethod
    def backward(ctx, *inputs):
        inputs = [i.data for i in inputs]
        inputs = [inputs[i:i + ctx.num_inputs]
                 for i in range(0, len(inputs), ctx.num_inputs)]
        results = comm.reduce_add_coalesced(inputs, ctx.target_gpus[0])
        outputs = comm.broadcast_coalesced(results, ctx.target_gpus)
        return (None,) + tuple([Variable(t) for tensors in outputs for t in tensors])

class Reduce(Function):
    @staticmethod
    def forward(ctx, *inputs):
        ctx.target_gpus = [inputs[i].get_device() for i in range(len(inputs))]
        inputs = sorted(inputs, key=lambda i: i.get_device())
        return comm.reduce_add(inputs)

    @staticmethod
    def backward(ctx, gradOutput):
        return Broadcast.apply(ctx.target_gpus, gradOutput)


class DataParallelModel(DataParallel):
    """Implements data parallelism at the module level.
    This container parallelizes the application of the given module by
    splitting the input across the specified devices by chunking in the
    batch dimension.
    In the forward pass, the module is replicated on each device,
    and each replica handles a portion of the input. During the backwards pass, gradients from each replica are summed into the original module.
    Note that the outputs are not gathered, please use compatible
    :class:`encoding.parallel.DataParallelCriterion`.
    The batch size should be larger than the number of GPUs used. It should
    also be an integer multiple of the number of GPUs so that each chunk is
    the same size (so that each GPU processes the same number of samples).
    Args:
        module: module to be parallelized
        device_ids: CUDA devices (default: all devices)
    Reference:
        Hang Zhang, Kristin Dana, Jianping Shi, Zhongyue Zhang, Xiaogang Wang, Ambrish Tyagi,
        Amit Agrawal. “Context Encoding for Semantic Segmentation.
        *The IEEE Conference on Computer Vision and Pattern Recognition (CVPR) 2018*
    Example::
        >>> net = encoding.nn.DataParallelModel(model, device_ids=[0, 1, 2])
        >>> y = net(x)
    """
    def gather(self, outputs, output_device):
        return outputs

    def replicate(self, module, device_ids):
        modules = super(DataParallelModel, self).replicate(module, device_ids)
        return modules


class DataParallelCriterion(DataParallel):
    """
    Calculate loss in multiple-GPUs, which balance the memory usage for
    Semantic Segmentation.
    The targets are splitted across the specified devices by chunking in
    the batch dimension. Please use together with :class:`encoding.parallel.DataParallelModel`.
    Reference:
        Hang Zhang, Kristin Dana, Jianping Shi, Zhongyue Zhang, Xiaogang Wang, Ambrish Tyagi,
        Amit Agrawal. “Context Encoding for Semantic Segmentation.
        *The IEEE Conference on Computer Vision and Pattern Recognition (CVPR) 2018*
    Example::
        >>> net = encoding.nn.DataParallelModel(model, device_ids=[0, 1, 2])
        >>> criterion = encoding.nn.DataParallelCriterion(criterion, device_ids=[0, 1, 2])
        >>> y = net(x)
        >>> loss = criterion(y, target)
    """
    def forward(self, inputs, *targets, **kwargs):
        # input should be already scatterd
        # scattering the targets instead
        if not self.device_ids:
            return self.module(inputs, *targets, **kwargs)
        targets, kwargs = self.scatter(targets, kwargs, self.device_ids)
        if len(self.device_ids) == 1:
            return self.module(inputs, *targets[0], **kwargs[0])
        replicas = self.replicate(self.module, self.device_ids[:len(inputs)])
        outputs = _criterion_parallel_apply(replicas, inputs, targets, kwargs)
        return Reduce.apply(*outputs) / len(outputs)


def _criterion_parallel_apply(modules, inputs, targets, kwargs_tup=None, devices=None):
    assert len(modules) == len(inputs)
    assert len(targets) == len(inputs)
    if kwargs_tup:
        assert len(modules) == len(kwargs_tup)
    else:
        kwargs_tup = ({},) * len(modules)
    if devices is not None:
        assert len(modules) == len(devices)
    else:
        devices = [None] * len(modules)

    lock = threading.Lock()
    results = {}
    if torch_ver != "0.3":
        grad_enabled = torch.is_grad_enabled()

    def _worker(i, module, input, target, kwargs, device=None):
        if torch_ver != "0.3":
            torch.set_grad_enabled(grad_enabled)
        if device is None:
            device = get_a_var(input).get_device()
        try:
            with torch.cuda.device(device):
                output = module(*(input + target), **kwargs)
            with lock:
                results[i] = output
        except Exception as e:
            with lock:
                results[i] = e

    if len(modules) > 1:
        threads = [threading.Thread(target=_worker,
                                    args=(i, module, input, target,
                                          kwargs, device),)
                   for i, (module, input, target, kwargs, device) in
                   enumerate(zip(modules, inputs, targets, kwargs_tup, devices))]

        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()
    else:
        _worker(0, modules[0], inputs[0], kwargs_tup[0], devices[0])

    outputs = []
    for i in range(len(inputs)):
        output = results[i]
        if isinstance(output, Exception):
            raise output
        outputs.append(output)
    return outputs